import asyncio
from typing import Optional, Union

from copy import copy

from shimoku.api.resources.app import App
from shimoku.api.resources.universe import Universe
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.file import File
from shimoku.api.resources.activity_template import ActivityTemplate

from shimoku.exceptions import AIFunctionError, ShimokuFileError, ActivityError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging
logger = logging.getLogger(__name__)


def check_app_is_set(self: Union['AILayer', 'AIFunctionMethods']):
    """ Check that the app is set """
    if self._app is None:
        log_error(logger, 'Menu path not set. Please use set_menu_path() method first.', AttributeError)


def get_model_metadata(model: File) -> dict:
    """ Get the metadata of a model
    :param model: Model
    :return: The metadata of the model
    """
    metadata: dict = copy(model['metadata'])
    for tag in model['tags']:
        if tag.startswith('creator_ai_function_version:'):
            metadata['creator_ai_function_version'] = tag[len('creator_ai_function_version:'):]
        elif tag.startswith('creator_ai_function_id'):
            metadata['creator_ai_function_id'] = tag[len('creator_ai_function_id:'):]
        elif tag.startswith('creator_ai_function'):
            metadata['creator_ai_function'] = tag[len('creator_ai_function:'):]
    return metadata


def get_output_file_metadata(file: File):
    """ Get the metadata of an output file
    :param file: File
    :return: The metadata of the output file
    """
    metadata: dict = copy(file['metadata'])
    for tag in file['tags']:
        if tag.startswith('creator_ai_function_version:'):
            metadata['creator_ai_function_version'] = tag[len('creator_ai_function_version:'):]
        elif tag.startswith('creator_ai_function:'):
            metadata['creator_ai_function'] = tag[len('creator_ai_function:'):]
        elif tag.startswith('model_name:'):
            metadata['model_name'] = tag[len('model_name:'):]
    return metadata


def get_output_file_name(
        activity_template: ActivityTemplate, file_name: str, run_id: str
) -> str:
    """ Get the name of an output file of a ai_function
    :param activity_template: Activity template of the ai_function
    :param file_name: Name of the file
    :param run_id: Id of the run
    :return: The name of the output file
    """
    name = activity_template['name']
    version = activity_template['version']
    return f"shimoku_generated_output_file_{name}_{version}_{run_id}_{file_name}"


def get_input_file_name(file_name: str):
    """ Get the name of an input file of a ai_function
    :param file_name: Name of the file
    :return: The name of the input file
    """
    return f"shimoku_generated_input_file_{file_name}"


async def check_and_get_model(app: App, model_name: str) -> File:
    """ Check that a model exists and get it
    :param app: App
    :param model_name: Name of the model
    """
    app_files = await app.get_files()
    for file in app_files:
        if 'shimoku_generated' not in file['tags'] or 'ai_model' not in file['tags']:
            continue
        metadata = get_model_metadata(file)
        if file['name'] == "shimoku_generated_model_" + model_name and metadata['model_name'] == model_name:
            return file
    log_error(logger, f"The model ({model_name}) does not exist", ShimokuFileError)


class AIFunctionMethods(ClassWithLogging):

    _module_logger = logger
    _use_info_logging = True

    def __init__(
            self, app: App, activity_template: ActivityTemplate, run_id: str
    ):
        self._app: App = app
        self._private_ai_function: ActivityTemplate = activity_template
        self._run_id: str = run_id

    async def _create_output_file(
            self, activity_template: ActivityTemplate, file_name: str,
            file: Union[bytes, tuple[bytes, dict]], run_id: str, model_name: Optional[str] = None
    ) -> str:
        """ Create an output file of a ai_function
        :param activity_template: Activity template of the ai_function
        :param file: File to be uploaded
        :param run_id: Id of the run
        :return: The uuid of the created file
        """
        metadata = {}
        if isinstance(file, tuple):
            file, metadata = file

        complete_file_name = get_output_file_name(activity_template, file_name, run_id)

        tags = ['shimoku_generated', 'ai_output_file',
                f'creator_ai_function:{activity_template["name"]}',
                f'creator_ai_function_version:{activity_template["version"]}']
        if model_name is not None:
            tags.append(f'model_name:{model_name}')

        file = await self._app.create_file(
            name=complete_file_name, file_object=file, tags=tags,
            metadata={
                'file_name': file_name,
                'run_id': run_id,
                **metadata
            }
        )
        logger.info(f'Created output file ({file_name})')
        return file['id']

    async def create_model(
            self, model_name: str, model: bytes, metadata: Optional[dict] = None
    ):
        """ Create a model
        :param model_name: Name of the model
        :param model: Model to be uploaded
        :param metadata: Metadata of the model
        """
        if not model_name or not isinstance(model_name, str):
            log_error(logger, 'Model name has to be a non-empty string', ValueError)

        file_name = f"shimoku_generated_model_{model_name}"

        check_app_is_set(self)
        await self._app.create_file(
            name=file_name, file_object=model,
            tags=[
                'shimoku_generated', 'ai_model',
                f"creator_ai_function:{self._private_ai_function['name']}",
                f"creator_ai_function_version:{self._private_ai_function['version']}",
                f"creator_ai_function_id:{self._private_ai_function['id']}"
            ],
            metadata={
                'model_name': model_name,
                'run_id': self._run_id,
                **(metadata or {})
            }
        )
        logger.info(f'Created model ({model_name})')

    async def get_model(
            self, model_name: str
    ) -> tuple[bytes, dict]:
        """ Get a model
        :param model_name: Name of the model
        :return: The model if it exists
        """
        check_app_is_set(self)
        model_file: File = await check_and_get_model(self._app, model_name)
        return await self._app.get_file_object(model_file['id']), get_model_metadata(model_file)

    async def create_output_files(
            self, files: dict[str, Union[bytes, tuple[bytes, dict]]], model_name: Optional[str] = None,
    ):
        """ Create output files of a ai_function
        :param files: Files to be uploaded
        :param model_name: Name of the model used
        """
        check_app_is_set(self)
        if model_name is not None:
            await check_and_get_model(self._app, model_name)
        else:
            model_name = ''

        await asyncio.gather(*[
            self._create_output_file(self._private_ai_function, file_name, file, self._run_id, model_name)
            for file_name, file in files.items()
        ])


async def get_universe_api_key(
    universe: Universe,
    universe_api_key: Optional[str]
) -> str:
    """ Get the universe API key or create one if none exists
    :param universe: Universe
    :param universe_api_key: Universe API key
    :return: Universe API key
    """
    if universe_api_key is None:
        universe_api_keys = await universe.get_universe_api_keys()
        if len(universe_api_keys) == 0:
            return (await universe.create_universe_api_key('AI functions key'))['id']
        return universe_api_keys[0]['id']
    return universe_api_key


async def get_activity_template(
    universe: Universe,
    name: str, version: Optional[str]
) -> ActivityTemplate:
    """ Get an activity template
    :param universe: Universe
    :param name: Name of the activity template
    :param version: Version of the activity template
    :return: Activity template
    """
    if version is not None:
        activity_template = await universe.get_activity_template(name_version=(name, version))
    else:
        activity_templates = await universe.get_activity_templates()
        activity_templates = [activity_template for activity_template in activity_templates
                              if activity_template['name'] == name]
        activity_template = activity_templates[-1] if len(activity_templates) > 0 else None
    if activity_template is None:
        log_error(
            logger,
            f"The AI function ({name} {'v' + (version if version is not None else '')}) does not exist",
            AIFunctionError
        )
    return activity_template


async def get_activity_from_template(
    app: App,
    activity_template: ActivityTemplate,
    universe_api_key: Optional[str] = None,
    create_if_not_exists: bool = True
) -> Activity:
    """ Get an activity from an activity template
    :param app: App
    :param activity_template: Activity template
    :param universe_api_key: Universe API key
    :param create_if_not_exists: Create the activity if it does not exist
    :return: Activity
    """
    universe_api_key = await get_universe_api_key(activity_template.parent, universe_api_key)

    name = activity_template['name']
    activity_name = 'shimoku_generated_activity_' + name

    activity: Optional[Activity] = await app.get_activity(name=activity_name)
    if activity is None:
        if not create_if_not_exists:
            log_error(logger, f"The activity for the AI function ({name}) does not exist", AIFunctionError)
        activity: Activity = await app.create_activity(
            name=activity_name,
            template_id=activity_template['id'],
            universe_api_key=universe_api_key
        )
        logger.info(f'Created activity for ai_function {name}')

    return activity


async def fill_input_for_run_metadata(
    key: str, value: any, args: dict, files: dict, input_settings: dict, app_files: list[File]
):
    """ Fill the files and args of a run metadata
    :param key: Key of the input
    :param value: Value of the input
    :param args: Args of the run metadata
    :param files: Files of the run metadata
    :param input_settings: Settings of the input
    :param app_files: Files of the app
    """
    if key not in input_settings:
        return
    if input_settings[key]['datatype'] == 'file':
        filtered_files: list[File] = [file for file in app_files if file['id'] == value]
        if len(filtered_files) > 1:
            log_error(logger, f"More than one file with ID {value} found", ValueError)
        elif len(filtered_files) == 0:
            value = '# file not found #'
        else:
            file = filtered_files[0]
            value = file['metadata']
        files[key] = value
    else:
        args[key] = value


async def fill_run_metadata(
    app: App, run_metadata: dict, run_id: str, ai_function_name: str, ai_function_version: str
):
    """ Fill the metadata of a run
    :param app: App
    :param run_metadata: Metadata of the run to fill
    :param run_id: ID of the run to fill
    :param ai_function_name: Name of the ai_function used
    :param ai_function_version: Version of the ai_function used
    """
    activity_template: ActivityTemplate = await get_activity_template(
        app.parent.parent, ai_function_name, ai_function_version
    )
    activity: Activity = await get_activity_from_template(app, activity_template, create_if_not_exists=False)
    activity_runs: list[Activity.Run] = await activity.get_runs()
    runs = [run for run in activity_runs if run['id'] == run_id]
    if len(runs) > 1:
        log_error(logger, f"More than one run with ID {run_id} found", ValueError)
    if len(runs) == 0:
        run_metadata['input'] = '# Run not found #'
        return
    run: Activity.Run = runs[0]
    input_settings = activity_template['inputSettings']
    args = {}
    files = {}
    await asyncio.gather(
        *[fill_input_for_run_metadata(k, v, args, files, input_settings, await app.get_files())
          for k, v in run['settings'].items()]
    )
    run_metadata['input']['args'] = args
    run_metadata['input']['files'] = files


async def check_and_store_output_file(
    file: File, files_by_run_id: list[dict], run_id: Optional[str] = None,
    model_metadata: Optional[dict] = None, activity_template: Optional[ActivityTemplate] = None,
    file_name: Optional[str] = None, get_objects: bool = False
):
    """ Check if a file is an output file then add it to the list of output files
    :param model_metadata: Metadata of the model used
    :param activity_template: ai_function used
    :param file: File to check
    :param files_by_run_id: dictionary of output files by run ID
    :param run_id: ID of the run to check
    :param file_name: Name of the file to check
    :param get_objects: Get the objects of the files instead of their metadata
    """
    if 'shimoku_generated' not in file['tags'] or 'ai_output_file' not in file['tags']:
        return

    output_file_metadata: dict = {}
    aux_output_file_metadata: dict = get_output_file_metadata(file)
    output_file_metadata.update(aux_output_file_metadata)

    output_file_name: str = output_file_metadata.pop('file_name')
    output_file_model_name: Optional[str] = output_file_metadata.get('model_name')
    ai_function_name: str = output_file_metadata.pop('creator_ai_function')
    ai_function_version: str = output_file_metadata.pop('creator_ai_function_version')

    if ((activity_template and
         (ai_function_name != activity_template['name'] or ai_function_version != activity_template['version'])) or
            (model_metadata and output_file_model_name != model_metadata['model_name']) or
            (file_name and output_file_name != file_name)):
        return

    output_file_run_id: str = output_file_metadata.pop('run_id')
    if run_id and output_file_run_id != run_id:
        return

    run_metadata = {'run_id': output_file_run_id}
    if output_file_run_id not in [run['run_id'] for run in files_by_run_id]:
        if not model_metadata and output_file_model_name is not None:
            run_metadata.update({
                'model_name': output_file_model_name
            })
        if not activity_template:
            run_metadata.update({
                'ai_function_name': ai_function_name,
                'ai_function_version': ai_function_version
            })
        run_metadata.update({
            'input': {'args': [], 'files': []},
            'output_files': {}
        })
        files_by_run_id.append(run_metadata)
    else:
        run_metadata = [run_d for run_d in files_by_run_id if run_d['run_id'] == output_file_run_id][0]

    if get_objects:
        output_file_metadata['object'] = await file.parent.get_file_object(file['id'])

    output_file_metadata = {k: v for k, v in output_file_metadata.items()
                            if k not in ['creator_ai_function', 'creator_ai_function_version']}
    run_metadata['output_files'][output_file_name] = output_file_metadata

    await fill_run_metadata(
        file.parent,
        run_metadata,
        output_file_run_id,
        ai_function_name,
        ai_function_version
    )


async def get_runs_message(
    runs: list[Activity.Run],
    wf_input_settings: Union[dict, list]
) -> list[str]:
    """ Get the output from a list of runs
    :param runs: list of runs
    :param wf_input_settings: Input settings of the ai_function/s
    :return: Message of the run
    """
    message = []
    # No need to use 'gather' here because logs are already cached from ordering the runs
    for i, run in enumerate(runs[::-1]):
        if isinstance(wf_input_settings, list):
            input_settings = wf_input_settings[i]
        else:
            input_settings = wf_input_settings
        run_settings = {k: v for k, v in run['settings'].items()
                        if k in input_settings and input_settings[k]['datatype'] != 'file'}
        message.extend([
            '',
            f' - Run {run["id"]}:',
            f'   Settings: {", ".join([f"{key}: {value}" for key, value in run_settings.items()])}',
            f'   Logs:',
            *[f'     - {log["severity"]} | {log["message"]}, at {log["dateTime"]}'
              for log in (await run.get_logs())]
        ])

    message.extend(['', "///////////////////////////////////////////////////", ''])
    return message


class AILayer(ClassWithLogging):

    _module_logger = logger
    _use_info_logging = True

    def __init__(
            self, access_token: str, universe: Universe, app: App,
    ):
        self._app: App = app
        self._universe: Universe = universe
        self._access_token: str = access_token
        self._private_ai_function_and_run_id: Optional[tuple[ActivityTemplate, str]] = None
        if app is not None and universe['id'] != app.parent.parent['id']:
            log_error(
                logger,
                f"Menu path ({str(app)}) does not belong to the universe ({str(universe)})",
                AIFunctionError
            )

    async def _check_run_id_exists(self, activity_template: ActivityTemplate, run_id: str):
        """ Check if a run ID exists for the activity of a ai_function
        :param activity_template: Activity template to check
        :param run_id: ID of the run to check
        """
        activity: Activity = await get_activity_from_template(self._app, activity_template, create_if_not_exists=False)
        if not await activity.get_run(run_id):
            log_error(
                logger,
                f"Run ({run_id}) does not exist for the activity ({activity_template['name']})",
                ActivityError
            )

    async def check_for_private_access(self, ai_function_id: str, run_id: str):
        """ Check if the credentials allow for private ai_function methods """
        universe_api_keys = [uak['id'] for uak in await self._universe.get_universe_api_keys()]
        if self._access_token not in universe_api_keys:
            log_error(
                logger,
                f"The access key must be a universe API key to use private AI function methods",
                AIFunctionError
            )

        activity_template: ActivityTemplate = await self._universe.get_activity_template(ai_function_id)

        check_app_is_set(self)
        await self._check_run_id_exists(activity_template, run_id)

        self._private_ai_function_and_run_id = (activity_template, run_id)

    def get_private_ai_function_methods(self) -> AIFunctionMethods:
        """ Get the private ai_function methods if the credentials allow for it """
        if not self._private_ai_function_and_run_id:
            log_error(
                logger,
                "Credentials must be set to use private AI function methods, "
                "please call check_for_private_access first",
                AIFunctionError
            )
        activity_template, run_id = self._private_ai_function_and_run_id
        return AIFunctionMethods(self._app, activity_template, run_id)

    async def delete_model(self, model_name: str):
        """ Delete a model
        :param model_name: Name of the model
        """
        check_app_is_set(self)
        model_file: File = await check_and_get_model(self._app, model_name)
        await self._app.delete_file(model_file['id'])
        logger.info(f'Deleted model ({model_name})')

    async def get_last_executions_with_logs(
            self, ai_function: str, version: Optional[str] = None, how_many: int = 1
    ):
        """ Get the logs of the executions of a ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param how_many: Number of executions to get
        :return: The logs of the ai_function
        """
        activity_template: ActivityTemplate = await get_activity_template(self._universe, ai_function, version)
        activity: Activity = await get_activity_from_template(self._app, activity_template, create_if_not_exists=False)
        runs: list[Activity.Run] = await activity.get_runs(how_many)
        return [run.cascade_to_dict() for run in runs]

    async def get_output_files_by_model(
            self, model_name: str, file_name: Optional[str] = None, get_objects: bool = False
    ):
        """ Get output files for the executions of ai_functions with a given model
        :param model_name: Name of the model to use
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        :return: dictionary of output files by the execution identifier
        """
        app_files: list[File] = await self._app.get_files()
        model_file: File = await check_and_get_model(self._app, model_name)
        model_metadata = get_model_metadata(model_file)

        files_by_run_id: list[dict] = []
        await asyncio.gather(*[
            check_and_store_output_file(
                file, files_by_run_id, model_metadata=model_metadata,
                file_name=file_name, get_objects=get_objects
            )
            for file in app_files
        ])
        return files_by_run_id

    async def get_output_files_by_ai_function(
            self, ai_function: str, version: Optional[str] = None,
            file_name: Optional[str] = None, get_objects: bool = False
    ):
        """ Get output files for a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :return: The output files if they exist
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        """
        app_files: list[File] = await self._app.get_files()
        activity_template: ActivityTemplate = await get_activity_template(self._universe, ai_function, version)

        files_by_run_id: list[dict] = []
        await asyncio.gather(*[
            check_and_store_output_file(
                file, files_by_run_id, activity_template=activity_template,
                file_name=file_name, get_objects=get_objects
            )
            for file in app_files
        ])
        return files_by_run_id

    async def get_output_file_objects(
            self, run_id: str, file_name: Optional[str] = None
    ) -> Union[tuple[bytes, dict], dict[str, tuple[bytes, dict]]]:
        """ Get an output file object for a given ai_function and run id
        :param run_id: ID of the executed run
        :param file_name: Name of the file to get
        """
        app_files: list[File] = await self._app.get_files()
        files_by_run_id: list[dict] = []
        await asyncio.gather(*[
            check_and_store_output_file(file, files_by_run_id, run_id, file_name=file_name, get_objects=True)
            for file in app_files
        ])
        if len(files_by_run_id) == 0:
            log_error(logger, f"No output files found for run ({run_id})", AIFunctionError)

        output_files_by_name = files_by_run_id[0]['output_files']

        if file_name is None:
            return_dict = {}
            for file_name, file_dict in output_files_by_name.items():
                return_dict[file_name] = file_dict['object'], {k: v for k, v in file_dict.items() if k != 'object'}
            return return_dict

        if file_name not in output_files_by_name:
            log_error(logger, f'No output file ({file_name}) found for run ({run_id})', AIFunctionError)

        file_dict = output_files_by_name[file_name]
        return file_dict['object'], {k: v for k, v in file_dict.items() if k != 'object'}

    async def _create_input_file(
            self, file_name: str, file: Union[any, tuple[any, dict]], force_overwrite: bool = False
    ):
        """ Create an output file of an ai_function
        :param file_name: Name of the file to create
        :param file: File to be uploaded
        :param force_overwrite: Overwrite the file if it already exists
        """
        metadata = {}
        if isinstance(file, tuple):
            file, metadata = file

        complete_file_name = get_input_file_name(file_name)
        if not force_overwrite and await self._app.get_file(name=complete_file_name) is not None:
            log_error(
                logger,
                f'Input file ({file_name}) already exists, if you want to overwrite it, use force_overwrite=True',
                AIFunctionError
            )
        await self._app.create_file(
            name=complete_file_name, file_object=file,
            tags=['shimoku_generated', 'ai_input_file'],
            metadata={
                'file_name': file_name,
                **metadata
            },
        )
        logger.info(f'Created input file {file_name}')

    async def _get_input_file(self, file_name: str) -> str:
        """ Get the ID of an input file
        :param file_name: Name of the file to get
        :return: ID of the file
        """
        file_name = get_input_file_name(file_name)
        input_file: Optional[File] = await self._app.get_file(name=file_name)

        if input_file is None:
            log_error(logger, f"Input file ({file_name}) not found", AIFunctionError)

        if input_file['tags'] != ['shimoku_generated', 'ai_input_file'] or 'file_name' not in input_file['metadata']:
            log_error(logger, f"File ({file_name}) is not an input file", AIFunctionError)

        return input_file['id']

    async def create_input_files(self, input_files: dict, force_overwrite: bool = False):
        """ Create input files for a ai_function
        :param input_files: dictionary of input files to create
        :param force_overwrite: Force the overwrite of the files
        """
        check_app_is_set(self)
        await asyncio.gather(*[
            self._create_input_file(file_name, file, force_overwrite)
            for file_name, file in input_files.items()
        ])

    async def _get_available_input_files(self) -> list[File]:
        """ Get available input files for a ai_function """
        app_files: list[File] = await self._app.get_files()
        input_files = [
            file for file in app_files if 'shimoku_generated' in file['tags'] and 'ai_input_file' in file['tags']
        ]
        return sorted(input_files, key=lambda x: x['metadata']['file_name'])

    async def get_available_input_files(self) -> list[dict]:
        """ Get available input files for a ai_function """
        check_app_is_set(self)
        return [file['metadata'] for file in await self._get_available_input_files()]

    async def delete_input_file(self, file_name: str):
        """ Delete an input file
        :param file_name: Name of the file to delete
        """
        check_app_is_set(self)
        file_id = await self._get_input_file(file_name)
        await self._app.delete_file(file_id)
        logger.info(f"Deleted input file {file_name}")

    async def generic_execute(
            self, ai_function: str, version: Optional[str] = None,
            universe_api_key: Optional[str] = None, **params
    ) -> str:

        """ Execute a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param universe_api_key: API key of the universe
        :param params: Parameters to be passed to the ai_function
        """
        async def check_params(
            _activity_template: ActivityTemplate, _params: dict
        ):
            """ Check the parameters passed to the ai_function, and create input files if necessary
            :param _activity_template: Activity template of the ai_function
            :param _params: Parameters to be passed to the ai_function
            """
            input_settings = _activity_template['inputSettings']
            if any(param not in input_settings for param in _params):
                log_error(
                    logger,
                    f"Unknown parameters for AI function ({_activity_template['name']}): "
                    f"{[param for param in _params if param not in input_settings]} \n"
                    f"The possible parameters are: {list(input_settings.keys())}",
                    AIFunctionError
                )
            for param_name, definition in input_settings.items():
                if param_name not in _params:
                    if definition['mandatory']:
                        log_error(
                            logger,
                            f"Missing parameter ({param_name}) for AI function ({_activity_template['name']}), "
                            f"the description of the missing parameter is: {definition['description']}",
                            AIFunctionError
                        )
                    else:
                        continue
                param_value = _params[param_name]
                param_definition_type = definition['datatype']
                if definition['datatype'] == 'file':
                    if not isinstance(param_value, str):
                        log_error(
                            logger,
                            f'The parameter ({param_name}) is expected to be a reference name to an input file',
                            AIFunctionError
                        )
                    _params[param_name] = await self._get_input_file(param_value)
                elif definition['datatype'] == 'model':
                    if not isinstance(param_value, str):
                        log_error(
                            logger,
                            f'The parameter ({param_name}) is expected to be a model name',
                            AIFunctionError
                        )
                    if not await check_and_get_model(self._app, param_value):
                        log_error(
                            logger,
                            f'The model ({param_value}) does not exist',
                            AIFunctionError
                        )
                elif str(type(param_value)) != f"<class '{param_definition_type}'>":
                    log_error(
                        logger,
                        f"Wrong type for parameter ({param_name}) for the AI function ({_activity_template['name']})"
                        f"the description of the missing parameter is: {definition['description']}\n"
                        f"Expected type: {param_definition_type}\n"
                        f"Provided type: {str(type(param_value))[8:-2]}",
                        AIFunctionError
                    )
        activity_template: ActivityTemplate = await get_activity_template(self._universe, ai_function, version)
        activity: Activity = await get_activity_from_template(self._app, activity_template, universe_api_key)
        await check_params(activity_template, params)
        run: Activity.Run = await activity.create_run(settings=params)
        logger.info(f'Response from execution trigger: {await run.trigger_webhook()}')
        return run['id']



from typing import Union, Optional, TYPE_CHECKING

import pandas as pd

from shimoku.exceptions import BentoboxError
from shimoku.plt.utils import create_normalized_name, ShimokuPalette
from shimoku_components_catalog.html_components import create_h1_title_with_modal

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


def check_for_bentobox(self: 'PlotLayer'):
    if self._bentobox_data:
        log_error(logger, 'This function groups a chart and a group of indicators with a bentobox,'
                          ' so it cannot be used inside another bentobox. Please'
                          ' pop out of the current bentobox before using this function.', BentoboxError)


def _call_inner_chart_function_with_parameters(
        self: 'PlotLayer', order: int, default_rows_size: int, default_cols_size: int,
        chart_parameters: dict, chart_function: Optional[callable] = None,
) -> any:
    if not chart_parameters or not isinstance(chart_parameters, dict):
        chart_parameters = {}

    restricted_parameters = ['order']

    for parameter in restricted_parameters:
        if parameter in chart_parameters:
            log_error(logger, f'Parameter {parameter} cannot be used in the inner chart', ValueError)

    chart_parameters.update(dict(
        order=order,
        padding='1,0,0,1' if 'padding' not in chart_parameters else chart_parameters['padding'],
        rows_size=default_rows_size if 'rows_size' not in chart_parameters else chart_parameters['rows_size'],
        cols_size=default_cols_size if 'cols_size' not in chart_parameters else chart_parameters['cols_size'],
    ))

    if chart_function is None:
        chart_function = self.bar

    return chart_function(**chart_parameters)


def infographics_text_bubble(
        self: 'PlotLayer', title: str, text: str, order: int, chart_parameters: dict,
        rows_size: int = 3, cols_size: int = 12, chart_function: Optional[callable] = None,
        background_url: Optional[str] = None, background_color: str = ShimokuPalette.BACKGROUND.value,
        bubble_location: str = 'top', image_url: Optional[str] = None, image_size: int = 100,
):
    if bubble_location not in ['top', 'bottom', 'left', 'right']:
        log_error(logger, f"Invalid location value '{bubble_location}'.", ValueError)

    chart_first = bubble_location in ['bottom', 'right']
    vertical = bubble_location in ['top', 'bottom']

    if image_url and vertical:
        log_error(logger, "The image cannot be displayed vertically.", ValueError)

    default_rows_size = int(10 * rows_size * (2 / 3)) - 2 * int(
        bubble_location == 'bottom') if vertical else 10 * rows_size
    default_cols_size = 23 if vertical else 15

    if image_url == 'default':
        image_url = 'https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63332131120af18c03d2b69a_APAME-about.svg'

    check_for_bentobox(self)
    self.set_bentobox(cols_size, rows_size)

    _call_inner_chart_function_with_parameters(
        self=self, order=order + int(not chart_first),
        default_rows_size=default_rows_size, default_cols_size=default_cols_size,
        chart_function=chart_function, chart_parameters=chart_parameters
    )

    chart_padding = chart_parameters['padding'].replace(' ', '').split(',')
    sides_padding = int(chart_padding[1]) + int(chart_padding[3])
    vertical_padding = int(chart_padding[0]) + int(chart_padding[2])

    html_cols_size = 22 - int(not vertical) * (chart_parameters['cols_size'] + sides_padding)
    html_rows_size = 10 * rows_size - int(vertical) * (chart_parameters['rows_size'] + vertical_padding)

    r_hash = create_normalized_name(self._get_chart_hash(order))

    html = (
            "<head>"
            f"<style>.block-text-{r_hash}{{padding:24px; "
            f"color: {ShimokuPalette.BLACK.value}; font-size: 16px;" +
            (f"background-color: {background_color}; " if background_url is None else "") +
            f"background-size:cover; background-repeat: no-repeat;background-image: url('{background_url}');"
            "background-position: center; border-radius:var(--border-radius-m);}</style>"
            "</head>"
            f"<div class='block-text-{r_hash}'>" +
            (f"<img src={image_url} width='{image_size}%'>"
             if image_url else "") +
            f"<h1>{title}</h1>"
            f"<p>{text}</p>"
            "</div>"
    )
    self.html(
        html=html,
        order=order + int(chart_first),
        rows_size=html_rows_size,
        cols_size=html_cols_size,
        padding=f'{int(bubble_location != "bottom")},1,1,1'
    )
    self.pop_out_of_bentobox()


def chart_and_modal_button(
        self: 'PlotLayer', order: int, chart_parameters: dict, button_modal: str, rows_size: int = 3,
        cols_size: int = 12, chart_function: Optional[callable] = None, button_label: str = 'Read more',
        button_side_text: str = "Click on the button to read more about this topic.",
):
    check_for_bentobox(self)
    self.set_bentobox(cols_size, rows_size)
    _call_inner_chart_function_with_parameters(
        self=self, order=order, default_rows_size=rows_size * 10 - 6, default_cols_size=22,
        chart_function=chart_function, chart_parameters=chart_parameters
    )

    chart_padding = chart_parameters['padding'].replace(' ', '').split(',')
    vertical_padding = int(chart_padding[0]) + int(chart_padding[2])

    button_rows_size = 10 * rows_size - chart_parameters['rows_size'] + vertical_padding - 2

    if cols_size > 3:
        html = (
            "<head>"
            f"<style>.rubik-text{{padding: 2%;"
            f"background-color: {ShimokuPalette.STRIPE_LIGHT.value};"
            f"background-size:cover; background-repeat: no-repeat;"
            "background-position: center; border-radius:var(--border-radius-m);"
            f"color: {ShimokuPalette.GRAY.value}; font-size: 16px;}}</style>"
            "</head>"
            "<div class='rubik-text'>"
            f"<p>{button_side_text}</p>"
            "</div>"
        )

        self.html(
            html=html,
            order=order + 1,
            rows_size=button_rows_size,
            cols_size=13 + int(0.5 * (cols_size - 4)),
            padding='0,1,0,1'
        )

    self.modal_button(
        order=order + 2, modal=button_modal, label=button_label,
        rows_size=button_rows_size, cols_size=2, padding='0,0,0,1',
    )
    self.pop_out_of_bentobox()


def chart_and_indicators(
        self: 'PlotLayer', order: int, chart_parameters: dict,
        indicators_groups: list[Union[pd.DataFrame, list[dict]]],
        indicators_parameters: Optional[dict] = None,
        chart_rows_size: int = 3, cols_size: int = 12,
        chart_function: Optional[callable] = None,
) -> int:
    check_for_bentobox(self)
    if indicators_parameters is None:
        indicators_parameters = {}

    self.set_bentobox(cols_size, chart_rows_size + len(indicators_groups))

    _call_inner_chart_function_with_parameters(
        self=self, order=order, default_rows_size=chart_rows_size * 14, default_cols_size=22,
        chart_function=chart_function, chart_parameters=chart_parameters
    )

    not_allowed_parameters = ['data', 'menu_path', 'order', 'bentobox_data']
    for parameter in not_allowed_parameters:
        if parameter in indicators_parameters:
            log_error(logger, f'Parameter {parameter} cannot be used in the inner indicators', ValueError)

    indicators_parameters['padding'] = indicators_parameters.get('padding', '0,0,0,1')
    indicators_parameters['rows_size'] = indicators_parameters.get('rows_size', 8)
    indicators_parameters['cols_size'] = indicators_parameters.get('cols_size', 23)

    order += 1
    for indicators_group in indicators_groups:
        order = self.indicator(data=indicators_group, order=order, **indicators_parameters)

    self.pop_out_of_bentobox()

    return order


def indicators_with_header(
        self: 'PlotLayer', order: int, indicators_groups: list[Union[pd.DataFrame, list[dict]]],
        title: str, indicators_parameters: Optional[dict] = None,
        subtitle: str = '', background_color: str = ShimokuPalette.PRIMARY.value,
        text_color: str = ShimokuPalette.BACKGROUND_PAPER.value, cols_size: int = 12,
        icon_url: str = 'https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63e3615716d4435d29e0b82c_Acurracy.svg',
) -> int:
    if indicators_parameters is None:
        indicators_parameters = {}
    return chart_and_indicators(
        self=self, order=order,
        chart_rows_size=1, cols_size=cols_size,
        chart_function=self.html,
        chart_parameters=dict(
            html=create_h1_title_with_modal(
                title=title, subtitle=subtitle, background_color=background_color,
                text_color=text_color, icon_url=icon_url
            ),
            padding='0,0,0,0',
            cols_size=24,
        ),
        indicators_groups=indicators_groups,
        indicators_parameters=indicators_parameters
    )


def line_with_summary(
        self: 'PlotLayer', order: int, data: Union[str, pd.DataFrame, list[dict]], title: str, x: str,
        y: Optional[Union[str, list[str]]] = None, description: str = '',
        value: Union[str, float, int] = '', rows_size: int = 2, cols_size: int = 2,
):
    check_for_bentobox(self)

    self.set_bentobox(cols_size, rows_size)

    _call_inner_chart_function_with_parameters(
        self=self, order=order, default_rows_size=(rows_size // 2 + rows_size % 2) * 12, default_cols_size=22,
        chart_function=self.area,
        chart_parameters={
            'data': data, 'x': x, 'y': y, 'variant': 'minimal',
            'option_modifications': {
                'xAxis': {
                    'axisLabel': {'show': True, 'showMaxLabel': True,
                                  'showMinLabel': True, 'interval': 10000000000000},
                    'boundaryGap': True
                },
                'yAxis': {
                    'axisLabel': {'show': False},
                },
                'grid': {'left': '1%', 'right': '1%', 'top': '1%', 'bottom': '11%'}
            }
        }
    )
    self.indicator(
        data={
            'title': title, 'description': description,
            'value': value, 'align': 'center', 'color': 'var(--color-black)'
        },
        order=order + 1,
        rows_size=(rows_size // 2 + rows_size % 2) * 8,
        cols_size=22,
        padding='0,0,0,1',
    )
    self.pop_out_of_bentobox()


def table_with_header(
    self: 'PlotLayer', order: int, data: Union[str, pd.DataFrame, list[dict]],
    title: str, table_parameters: Optional[dict] = None,
    subtitle: str = '', rows_size: int = 3, cols_size: int = 4, modal_title: Optional[str] = None,
    icon_url: str = 'https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/'
                    '63594ccf3f311a98d72faff7_suite-customer-b.svg',
    modal_text: str = "You can click the X in the corner or click the overlay to close this modal. "
                      "This is a nice way to show additional information.",
    text_color: str = ShimokuPalette.BACKGROUND_PAPER.value,
    background_color: str = ShimokuPalette.BLACK.value, modal_icon_color: str = ShimokuPalette.CHART_C1.value,
    modal_icon_hover_color: str = ShimokuPalette.PRIMARY_DARK.value,
):
    check_for_bentobox(self)
    self.set_bentobox(cols_size=cols_size, rows_size=rows_size)
    self.html(
        html=create_h1_title_with_modal(
            title=title, subtitle=subtitle, background_color=background_color,
            text_color=text_color, icon_url=icon_url, modal_title=modal_title,
            modal_text=modal_text, modal_icon_color=modal_icon_color,
            modal_icon_hover_color=modal_icon_hover_color
        ),
        order=order,
        cols_size=24, rows_size=1,
        padding='0,0,0,0',
    )

    if table_parameters is None:
        table_parameters = {}
    table_parameters['data'] = data
    table_parameters['padding'] = table_parameters.get('padding', '0,0,0,1')

    _call_inner_chart_function_with_parameters(
        self=self, order=order+1, default_rows_size=rows_size,
        default_cols_size=22, chart_function=self.table, chart_parameters=table_parameters
    )
    self.pop_out_of_bentobox()



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def sunburst_chart(
    self: 'PlotLayer', order: int, data: Union[str, dict, list],
    title: Optional[str] = None, padding: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a sunburst chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['xAxis']
    del common_options['yAxis']
    del common_options['toolbox']['feature']['magicType']
    if isinstance(data, dict):
        data = [data]
    series_options.update({
        'type': 'sunburst',
        'data': '#set_data#',
        'radius': [0, '95%'],
        'label': {
            'rotate': 'radial',
        }
    })
    common_options.update({
        'tooltip': {
            'trigger': 'item',
            'triggerOn': 'mousemove'
        },
        'series': series_options
    })
    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data, dump_whole=True),
        fields=['data'], options=common_options, order=order, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding,
        option_modifications=option_modifications
    )



from shimoku.api.resources.universe import Universe
from shimoku import ApiClient

import logging
from shimoku.execution_logger import ClassWithLogging
logger = logging.getLogger(__name__)


class UniversesLayer(ClassWithLogging):
    """
    Class used to interact with the API at the universe level
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self, api_client: ApiClient
    ):
        self._api_client = api_client

    async def create_universe_api_key(
        self, uuid: str, description: str
    ) -> dict:
        """
        Create a universe API key, this key has admin privileges.
        :param uuid: uuid of the universe
        :param description: description of the key
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return await universe.create_universe_api_key(description)

    async def get_universe_api_keys(
        self, uuid: str
    ) -> list[dict]:
        """
        Get the universe API keys.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return await universe.get_universe_api_keys()

    async def delete_universe_api_key(
        self, uuid: str, api_key_uuid: str
    ) -> bool:
        """
        Delete a universe API key.
        :param uuid: uuid of the universe
        :param api_key_uuid: uuid of the API key
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return await universe.delete_universe_api_key(api_key_uuid)

    async def get_universe_workspaces(
        self, uuid: str
    ) -> list[dict]:
        """
        Get the universe workspaces.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [b.cascade_to_dict() for b in await universe.get_businesses()]

    async def get_universe_activity_templates(
        self, uuid: str
    ) -> list[dict]:
        """
        Get the universe activity templates as a list of dictionaries.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [{
            'name': at['name'], 'description': at['description'],
            'min_run_interval': at['minRunInterval'],
            'input_settings': [{
                'name': name,
                'description': param['description'],
                'datatype': param['datatype']
            } for name, param in at['inputSettings'].items() if isinstance(param, dict)]
        } for at in await universe.get_activity_templates() if at['enabled']]



""""""
import unittest
from os import getenv
from typing import Dict, List
from shimoku.exceptions import WorkspaceError, CacheError, APIError

import unittest

from utils import initiate_shimoku

s = initiate_shimoku()

business_id: str = getenv('BUSINESS_ID')
mock: bool = getenv('MOCK') == 'TRUE'
playground: bool = getenv('PLAYGROUND') == 'TRUE'
BUSINESS_TEST_NAME = 'auto-test'


class TestBusinesses(unittest.TestCase):

    def test_get_business(self):
        result = s.workspaces.get_workspace(uuid=business_id)
        assert result

    def test_get_fake_business(self):
        if not mock and not playground:
            with self.assertRaises(APIError):
                s.workspaces.get_workspace(
                    uuid='this is a test',
                )

    def test_create_and_delete_business(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        business: Dict = s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)
        business_id_: str = business['id']

        assert len(business_id_) > 0

        business_from_db: Dict = (
            s.workspaces.get_workspace(
                uuid=business_id_,
            )
        )

        business_roles = s.workspaces.get_roles(name=BUSINESS_TEST_NAME)

        assert len(business_roles) == 4
        resources = ['DATA', 'DATA_EXECUTION', 'USER_MANAGEMENT', 'BUSINESS_INFO']
        for role in business_roles:
            assert role['role'] == 'business_read'
            assert role['permission'] == 'READ'
            assert role['resource'] in resources
            assert role['target'] == 'GROUP'
            resources.remove(role['resource'])

        assert {
            k: v
            for k, v in business_from_db.items()
            if k in [
                'id', 'name',
                'universe', '__typename',
            ]
        } == {
            k: v
            for k, v in business.items()
            if k in [
                'id', 'name',
                'universe', '__typename',
            ]
        }
        del business_from_db

        s.workspaces.delete_workspace(uuid=business_id_)

        if not mock:
            with self.assertRaises(APIError):
                s.workspaces.get_workspace(
                    uuid=business_id_,
                )

    def test_theme_customization(self):
        s.workspaces.update_workspace(
            uuid=business_id,
            theme={},
        )
        business: Dict = s.workspaces.get_workspace(uuid=business_id)

        assert business['theme'] == {}

        theme = {
            "palette": {
                "primary": {
                    "main": "#2222FF",
                },
                "background": {
                    "default": "#23232F",
                },
            },
            "typography": {
                "h1": {
                    "fontSize": "60px",
                },
            },
            "custom": {
                "radius": {
                    "xs": "0px",
                    "s": "0px",
                    "m": "0px",
                    "l": "0px",
                    "xl": "0px"
                },
                "logo": "https://assets-global.website-files.com/619f9fe98661d321dc3beec7"
                        "/621f5f09cd144b3dc06dc0fd_Logo-shimoku-white.svg",
            }
        }

        s.workspaces.update_workspace(uuid=business_id, theme=theme)

        business: Dict = s.workspaces.get_workspace(uuid=business_id)

        assert business['theme'] == theme

    def test_update_business(self):
        business_original: Dict = s.workspaces.get_workspace(uuid=business_id)

        business_name: str = business_original['name']
        new_business_name: str = f'{business_name}_changed'

        s.workspaces.update_workspace(
            uuid=business_id,
            new_name=new_business_name,
        )

        new_business: Dict = s.workspaces.get_workspace(uuid=business_id)

        business_changed: Dict = s.workspaces.get_workspace(name=new_business_name)
        assert {
            k: v
            for k, v in business_changed.items()
            if k in [
                'id', 'name',
                'universe', '__typename',
            ]
        } == {
            k: v
            for k, v in new_business.items()
            if k in [
                'id', 'name',
                'universe', '__typename',
            ]
        }
        assert new_business_name == business_changed['name']

        # Restore previous name
        s.workspaces.update_workspace(name=new_business_name, new_name=business_name)

        business_restored: Dict = s.workspaces.get_workspace(name=business_name)
        assert business_name == business_restored['name']

    def test_get_business_apps(self):
        s.set_workspace(uuid=business_id)
        how_many_apps = len(s.workspaces.get_workspace_menu_paths(uuid=business_id))
        s.menu_paths.get_menu_path(name='business_app_test')
        apps: List[Dict] = s.workspaces.get_workspace_menu_paths(uuid=business_id)
        assert len(apps) == how_many_apps + 1
        s.menu_paths.delete_menu_path(name='business_app_test')
        apps: List[Dict] = s.workspaces.get_workspace_menu_paths(uuid=business_id)
        assert len(apps) == how_many_apps

    def test_cant_create_business_with_similar_name(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)

        class SimilarBusinessCase(unittest.TestCase):
            def test_create_similar(self):
                with self.assertRaises(WorkspaceError):
                    s.workspaces.create_workspace(
                        name='auto-TEST',
                    )

            def test_update_similar(self):
                with self.assertRaises(WorkspaceError):
                    s.workspaces.update_workspace(
                        name=BUSINESS_TEST_NAME,
                        new_name='auto-TEST',
                    )
        sbc = SimilarBusinessCase()
        sbc.test_create_similar()
        sbc.test_update_similar()
        s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

    def test_cant_get_with_id_and_name(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        wid = s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)['id']

        with self.assertRaises(CacheError):
            s.set_workspace(uuid=wid, name=BUSINESS_TEST_NAME)



from typing import List, Dict, Optional, Type, Tuple, Union, TYPE_CHECKING
from copy import deepcopy

import pandas as pd
import asyncio

from shimoku.api.base_resource import Resource, ResourceCache
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.role import Role, create_role, get_role, get_roles, delete_role
from shimoku.api.resources.report import Report
from shimoku.api.resources.file import File
from shimoku.api.resources.data_set import DataSet, Mapping

from shimoku.utils import create_normalized_name, EventType
from shimoku.exceptions import MenuPathError, UniverseApiKeyError, ActivityTemplateError

if TYPE_CHECKING:
    from .business import Business

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class App(Resource):
    """ App resource class """

    resource_type = 'app'
    alias_field = 'name'
    plural = 'apps'

    def __init__(self, parent: 'Business', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[Dict] = None):

        normalized_name: Optional[str] = None
        if alias:
            normalized_name: str = create_normalized_name(alias)

        params = dict(
            name=alias,
            order=0,
            normalizedName=normalized_name,
            hideTitle=True,
            hidePath=False,
            showBreadcrumb=False,
            showHistoryNavigation=False,
        )

        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource,
                         children=[Role, Activity, Report, DataSet, File],
                         check_params_before_creation=['name'], params=params)

        self.currently_in_use: bool = False

    async def delete(self):
        """ Delete the app """
        if await self.check_for_shimoku_generated_files():
            log_error(
                logger,
                f'The menu path ({str(self)}) cannot be deleted, '
                f'because it contains files generated by the SDK. If you are sure you want to delete those files, '
                f'please use the method (menu_paths.delete_all_menu_path_files) '
                f'with the parameter (with_shimoku_generated) set to True',
                MenuPathError
            )
        if self.currently_in_use:
            log_error(logger, f'Menu path ({str(self)}) is currently in use and cannot be deleted', MenuPathError)
        return await self._base_resource.delete()

    async def update(self):
        """ Update the app """
        return await self._base_resource.update()

    # Activity methods
    async def _get_activity_template_id(
        self, template_id: Optional[str] = None, template_name_version: Optional[Tuple[str, str]] = None
    ) -> str:
        """
        Get the activity template id from the template id or name
        :param template_id: the template id
        :param template_name_version: the template name and version
        """
        activity_template = await self.parent.parent.get_activity_template(
            uuid=template_id, name_version=template_name_version
        )
        if activity_template:
            return activity_template['id']

        log_error(
            logger,
            'Activity template not found in the universe, please use an existing one.',
            ActivityTemplateError
        )

    async def create_activity(
        self, name: str, settings: Optional[Dict[str, str]] = None,
        template_id: Optional[str] = None, template_name_version: Optional[Tuple[str, str]] = None,
        template_mode: str = 'LIGHT', universe_api_key: str = ''
    ) -> Activity:
        """
        Create an activity, either from a template or from scratch
        :param name: Name of the activity
        :param settings: Settings of the activity
        :param template_id: UUID of the activity template
        :param template_name_version: Name and version of the activity template
        :param template_mode: Mode of the activity template
        :param universe_api_key: Universe API key to use
        """
        template_params_to_send = {}

        if template_id or template_name_version:
            template_id = await self._get_activity_template_id(template_id, template_name_version)
            universe_api_keys = await self.parent.parent.get_universe_api_keys()
            if universe_api_key not in [u['id'] for u in universe_api_keys]:
                log_error(logger, 'Universe API key not found in the universe, please use an existing one.',
                          UniverseApiKeyError)

            template_params_to_send = dict(
                activityTemplateWithMode=dict(
                    activityTemplateId=template_id,
                    mode=template_mode,
                ),
                universeApiKeyId=universe_api_key,
            )

        return await self._base_resource.create_child(
            Activity, alias=name, settings=settings, **template_params_to_send
        )

    async def update_activity(self, uuid: Optional[str] = None, name: Optional[str] = None, **params) -> Activity:
        if params.get('new_name') is not None:
            params['name'] = params.pop('new_name')
            params['new_alias'] = True
        return await self._base_resource.update_child(Activity, uuid=uuid, alias=name, **params)

    async def get_activity(self, uuid: Optional[str] = None, name: Optional[str] = None) -> Optional[Activity]:
        return await self._base_resource.get_child(Activity, uuid, name)

    async def get_activities(self) -> List[Activity]:
        return await self._base_resource.get_children(Activity)

    async def delete_activity(self, uuid: Optional[str] = None, name: Optional[str] = None):
        return await self._base_resource.delete_child(Activity, uuid, name)

    # Report methods
    def mock_create_report(self, report_class: Type[Report], r_hash: str, **params) -> Report:
        """ Creates a child of a given report class but doesn't add it to the cache.
        :param report_class: The class of the report to create.
        :param r_hash: The hash of the report to create.
        :param params: The parameters of the report to create.
        :return: The created resource.
        """
        report = report_class(parent=self)
        params = deepcopy(params)

        properties = params.pop('properties') if 'properties' in params else {}
        properties['hash'] = r_hash

        report.set_properties(**properties)
        report.set_params(**params)
        return report

    async def get_paths_in_order(self) -> List[str]:
        """ Gets all the paths of the app. They are stored in the reports. """
        reports = await self.get_reports()
        path_orders = [(report['path'], report['pathOrder']) for report in reports if report['pathOrder'] is not None]
        path_orders.sort(key=lambda x: x[1])
        result = []
        for path_order in path_orders:
            if path_order[0] not in result:
                result.append(path_order[0])
        return result

    async def create_report(self, report_class: Type[Report], r_hash: str, **params) -> Report:
        """ Creates a child of a given report class.
        :param report_class: The class of the report to create.
        :param r_hash: The hash of the report to create.
        :param params: The parameters of the report to create.
        :return: The created resource.
        """
        report_cache: ResourceCache = self._base_resource.children[Report]
        report = await report_cache.add(self.mock_create_report(report_class, r_hash, **params), alias=r_hash)
        return report

    async def get_report(self, uuid: Optional[str] = None, r_hash: Optional[str] = None) -> Optional[Report]:
        """ Gets a report.
        :param uuid: The UUID of the report to get.
        :param r_hash: The hash of the report to get.
        :return: The resource.
        """
        return await self._base_resource.get_child(Report, uuid, r_hash)

    async def get_reports(self) -> List[Report]:
        """ Gets all the reports of the app."""
        return await self._base_resource.get_children(Report)

    async def update_report(self, uuid: Optional[str] = None, r_hash: Optional[str] = None, **params):
        """ Updates a report.
        :param uuid: The UUID of the report to update.
        :param r_hash: The hash of the report to update.
        :param params: The parameters of the report to update.
        """
        report: Optional[Report] = await self.get_report(uuid, r_hash)
        if not report:
            logger.warning(f'Report {r_hash if r_hash else uuid} not found')
            return
        params = deepcopy(params)
        if 'properties' in params:
            if 'hash' in params['properties']:
                log_error(logger, 'Cannot update the hash of a component.', ValueError)
            report.set_properties(**params.pop('properties'))

        report.set_params(**params)
        await report.update()

    async def delete_report(self, uuid: Optional[str] = None, r_hash: Optional[str] = None):
        """ Deletes a report.
        :param uuid: The UUID of the report to delete.
        :param r_hash: The hash of the report to delete.
        """
        report: Optional[Report] = await self.get_report(uuid, r_hash)
        if not report:
            return
        await self._base_resource.delete_child(Report, uuid, r_hash)
        await self._base_resource.parent.create_event(EventType.REPORT_DELETED, {}, report['id'])

    # DataSet methods
    async def get_data_set(self, uuid: Optional[str] = None, name: Optional[str] = None, create_if_not_exists=True
                           ) -> Optional[DataSet]:
        """ Gets a dataset.
        :param uuid: The UUID of the dataset to get.
        :param name: The alias of the dataset to get.
        :param create_if_not_exists: If True, creates the dataset if it doesn't exist.
        :return: The dataset.
        """
        return await self._base_resource.get_child(DataSet, uuid, name, create_if_not_exists=create_if_not_exists)

    async def get_data_sets(self) -> List[DataSet]:
        """ Gets all the datasets of the app."""
        return await self._base_resource.get_children(DataSet)

    async def delete_data_set(self, uuid: Optional[str] = None, name: Optional[str] = None):
        """ Deletes a dataset.
        :param uuid: The UUID of the dataset to delete.
        :param name: The alias of the dataset to delete.
        """
        return await self._base_resource.delete_child(DataSet, uuid, name)

    async def append_data_to_data_set(
        self, data: Union[pd.DataFrame, List[Dict]],
        uuid: Optional[str] = None, name: Optional[str] = None,
        sort: Optional[Dict] = None, dump_whole: bool = False
    ) -> Tuple[Mapping, DataSet, Dict]:
        """ Appends data to a dataset. If the dataset doesn't exist, it is created.
        :param uuid: The UUID of the dataset to update.
        :param name: The alias of the dataset to update.
        :param data: The data to update the dataset with.
        :param sort: The sort to apply to the data.
        :param dump_whole: Whether to dump the whole data.
        :return: The dataset.
        """
        data_set: DataSet = await self.get_data_set(uuid, name)
        mapping, res_sort = await data_set.create_data_points(
            data_points=pd.DataFrame(data) if isinstance(data, list) and not dump_whole else data,
            sort=sort, dump_whole=dump_whole
        )
        return mapping, data_set, res_sort

    async def force_delete_data_set(self, uuid: Optional[str] = None, name: Optional[str] = None) -> bool:
        """ Deletes a dataset.
        :param uuid: The UUID of the dataset to delete.
        :param name: The alias of the dataset to delete.
        :return: Whether the dataset was deleted.
        """
        data_set = await self.get_data_set(uuid, name, create_if_not_exists=False)
        if not data_set:
            return False
        reports = await self.get_reports()
        report_data_sets_lists = await asyncio.gather(*[report.get_report_data_sets() for report in reports])
        delete_tasks = []
        for i, report_data_sets in enumerate(report_data_sets_lists):
            await asyncio.gather(*[reports[i].delete_report_dataset(rds['id'])
                                   for rds in report_data_sets if rds['dataSetId'] == data_set['id']])

        await asyncio.gather(*delete_tasks)
        await self.delete_data_set(uuid, name)
        return True

    async def delete_unused_data_sets(self, log: bool = False, data_set_ids: Optional[List[str]] = None):
        """ Deletes all the unused datasets of the app.
        :param log: Whether to log the deletion.
        :param data_set_ids: The ids of the datasets to delete.
        """
        all_reports = await self.get_reports()
        all_rds = await asyncio.gather(*[report.get_report_data_sets() for report in all_reports])
        all_datasets_in_use = [rds['dataSetId'] for _rds_list in all_rds for rds in _rds_list]
        all_datasets = await self.get_data_sets()
        dataset_ids_to_delete = [ds['id'] for ds in all_datasets
                                 if ds['id'] not in all_datasets_in_use
                                 and (ds['id'] in data_set_ids or data_set_ids is None)]
        await asyncio.gather(*[self._base_resource.delete_child(DataSet, ds_id) for ds_id in dataset_ids_to_delete])
        if log:
            logger.info(f'Deleted {len(dataset_ids_to_delete)} unused datasets from the menu path {str(self)}')

    # File methods
    async def check_for_shimoku_generated_files(self) -> bool:
        """ Checks whether the app has files generated by shimoku.
        :return: Whether the app has files generated by shimoku.
        """
        files = await self.get_files()
        return any(['shimoku_generated' in file['tags'] for file in files])

    async def create_file(
        self, name: str, file_object: bytes, tags: list, metadata: dict, overwrite: bool = True
    ) -> File:
        """ Creates a file.
        :param name: The file to create.
        :param file_object: The file object.
        :param tags: The tags of the file.
        :param metadata: The metadata of the file.
        :param overwrite: Whether to overwrite the file if it already exists.
        :return: The created file.
        """
        if overwrite and await self._base_resource.get_child(File, alias=name, create_if_not_exists=False):
            logger.info(f'Overwriting file {name}')
            await self.delete_file(name=name)

        name = await self._base_resource.create_child(
            File, alias=name, contentType='text/csv',
            tags=tags, metadata=metadata
        )
        logger.info(f'Uploading file {str(name)}')
        await self.api_client.request(
            'PUT', name['url'],
            body=file_object,
            headers={'content-type': 'text/csv'},
            to_tazawa=False
        )

        return name

    async def delete_file(self, uuid: Optional[str] = None, name: Optional[str] = None):
        """ Deletes a file.
        :param uuid: The UUID of the file to delete.
        :param name: The name of the file to delete.
        """
        return await self._base_resource.delete_child(File, uuid, name)

    async def get_file(self, uuid: Optional[str] = None, name: Optional[str] = None) -> Optional[File]:
        """ Gets a file.
        :param uuid: The UUID of the file to get.
        :param name: The alias of the file to get.
        :return: The file.
        """
        return await self._base_resource.get_child(File, uuid, name)

    async def get_files(self) -> List[File]:
        """ Gets all the files of the app.
        :return: The files.
        """
        return await self._base_resource.get_children(File)

    async def get_file_object(self, uuid: Optional[str] = None, name: Optional[str] = None) -> Optional[bytes]:
        """ Gets a file object.
        :param uuid: The UUID of the file to get.
        :param name: The alias of the file to get.
        :return: The file object.
        """
        file = await self.get_file(uuid, name)
        if not file:
            logger.warning(f'Could not get file {name}')
            return None

        await file.get()
        logger.info(f'Downloading file {str(file)}')
        return await self.api_client.request('GET', file['url'], to_tazawa=False)

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role



from typing import TYPE_CHECKING, Optional, Union
from pandas import DataFrame

from copy import deepcopy

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


def thin_variant(y: list[str], series_options: dict):
    len_y = len(y)
    dot_series_options = deepcopy(series_options)
    series_options['barWidth'] = 3
    series_options['barGap'] = '-100%'
    series_options['label'] = {'show': False}
    if len_y > 1:
        series_options['itemStyle']['opacity'] = 0.75
    series_options = len_y * [series_options]
    dot_series_options['type'] = 'scatter'
    dot_series_options['itemStyle']['opacity'] = 1
    dot_series_options['itemStyle']['borderWidth'] = 0
    dot_series_options['label'] = {'position': 'top', 'color': 'inherit'}
    series_options.append(dot_series_options)
    y = y + y
    return y, series_options


def shadow_variant(series_options: dict):
    series_options['showBackground'] = True
    series_options['backgroundStyle'] = {'color': 'rgba(180, 180, 180, 0.2)'}
    series_options['label'] = {'position': 'top', 'color': 'inherit'}
    return series_options


def horizontal_clean_variant(common_options: dict):
    common_options['xAxis'][0].update({
        'position': 'top',
        'axisLabel': {
            'rotate': -45,
        }
    })
    return common_options


def interpret_variant(variant: Optional[str]) -> tuple[Optional[str], Optional[str]]:
    if not variant:
        return None, None
    variant = variant.split(' ')
    if len(variant) == 1:
        if variant[0] in ['thin', 'shadow']:
            return None, variant[0]
        elif variant[0] in ['clean', 'minimal']:
            return variant[0], None
        raise ValueError(f'Invalid variant: {variant}')
    elif len(variant) == 2 and variant[0] in ['clean', 'minimal']:
        return variant[0], variant[1]
    else:
        raise ValueError(f'Invalid variant: {variant}')


async def bar_chart(
        self: 'PlotLayer', data: Union[str, DataFrame, list[dict]],
        order: int, x: str, y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, title: Optional[str] = None,
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
        variant: Optional[str] = None
):
    """ Create a bar chart """
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options['type'] = 'bar'
    clean, variant = interpret_variant(variant)

    if variant == 'thin':
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
    elif variant == 'shadow':
        series_options = shadow_variant(series_options)

    await self._create_trend_chart(
        echart_options=get_common_echart_options(), variant=clean,
        series_options=series_options, option_modifications=option_modifications, axes=x, values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )


async def stacked_bar_chart(
        self: 'PlotLayer', data: Union[str, DataFrame, list[dict]],
        order: int, x: str, y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, title: Optional[str] = None,
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
        variant: Optional[str] = None
):
    """ Create a stacked bar chart """
    series_options = get_common_series_options()
    series_options.update({
        'type': 'bar',
        'areaStyle': {},
        'stack': 'stack',
        'itemStyle': {'borderRadius': [0, 0, 0, 0]},
    })
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    clean, variant = interpret_variant(variant)
    if variant == 'thin':
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        len_y = len(y)
        y, series_options = thin_variant(y, series_options)
        series_options[-1]['stack'] = 'dot_stack'
        series_options[-1]['symbolSize'] = 5
        series_options[-1]['label']['position'] = 'left'
        series_options = series_options[:len_y] + [series_options[-1]] * (len_y - 1) + [deepcopy(series_options[-1])]
        series_options[-1]['label']['position'] = 'top'
        del series_options[-1]['symbolSize']
    else:
        len_y = len(y) if y else len(data_mapping_to_tuples.keys()) - 1
        if len_y > 1:
            series_options = [series_options] * (len_y - 1)
            series_options.append(deepcopy(series_options[-1]))
        else:
            series_options = [series_options]
        if variant == 'shadow':
            series_options[-1] = shadow_variant(series_options[-1])
        series_options[-1]['itemStyle']['borderRadius'] = [9, 9, 0, 0]

    await self._create_trend_chart(
        echart_options=get_common_echart_options(), variant=clean,
        axes=x, values=y, data_mapping_to_tuples=data_mapping_to_tuples,
        series_options=series_options, option_modifications=option_modifications,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )


async def zero_centered_bar_chart(
        self: 'PlotLayer', data: Union[str, DataFrame, list[dict]],
        order: int, x: str, y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, title: Optional[str] = None,
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
        variant: Optional[str] = None
):
    """ Create a zero centered bar chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options['type'] = 'bar'
    common_options['xAxis'][0]['type'] = 'value'
    series_options['itemStyle']['borderRadius'] = [0, 0, 0, 0]
    del common_options['xAxis'][0]['data']
    common_options['yAxis'][0].update({
        'type': 'category',
        'data': '#set_data#',
    })
    clean, variant = interpret_variant(variant)
    if variant == 'thin':
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
    elif variant == 'shadow':
        series_options = shadow_variant(series_options)

    await self._create_trend_chart(
        echart_options=common_options if not clean else horizontal_clean_variant(common_options),
        series_options=series_options, variant=clean,
        option_modifications=option_modifications, axes=x, values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )


async def horizontal_bar_chart(
        self: 'PlotLayer', data: Union[str, DataFrame, list[dict]],
        order: int, x: str, y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, title: Optional[str] = None,
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
        variant: Optional[str] = None
):
    """ Create a horizontal bar chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options['type'] = 'bar'
    common_options['xAxis'][0]['type'] = 'value'
    series_options['itemStyle']['borderRadius'] = [0, 9, 9, 0]
    del common_options['xAxis'][0]['data']
    common_options['yAxis'][0].update({
        'type': 'category',
        'data': '#set_data#',
    })
    clean, variant = interpret_variant(variant)
    if variant == 'thin':
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
        series_options[-1]['label']['position'] = 'right'
    elif variant == 'shadow':
        series_options = shadow_variant(series_options)
        series_options['label']['position'] = 'right'

    await self._create_trend_chart(
        echart_options=common_options if not clean else horizontal_clean_variant(common_options),
        series_options=series_options, variant=clean,
        option_modifications=option_modifications, axes=x, values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )


async def stacked_horizontal_bar_chart(
        self: 'PlotLayer', data: Union[str, DataFrame, list[dict]],
        order: int, x: str, y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, title: Optional[str] = None,
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
        variant: Optional[str] = None
):
    """ Create a stacked horizontal bar chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update({
        'type': 'bar',
        'areaStyle': {},
        'stack': 'stack',
        'itemStyle': {'borderRadius': [0, 0, 0, 0]},
    })
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    clean, variant = interpret_variant(variant)
    if variant == 'thin':
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        len_y = len(y)
        y, series_options = thin_variant(y, series_options)
        series_options[-1]['stack'] = 'dot_stack'
        series_options[-1]['symbolSize'] = 5
        series_options[-1]['label']['position'] = 'top'
        series_options = series_options[:len_y] + [series_options[-1]] * (len_y - 1) + [deepcopy(series_options[-1])]
        series_options[-1]['label']['position'] = 'right'
        del series_options[-1]['symbolSize']
    else:
        len_y = len(y) if y else len(data_mapping_to_tuples.keys()) - 1
        series_options = [series_options] * (len_y - 1)
        series_options.append(deepcopy(series_options[-1]))
        series_options[-1]['itemStyle']['borderRadius'] = [0, 9, 9, 0]
        if variant == 'shadow':
            series_options[-1] = shadow_variant(series_options[-1])
            series_options[-1]['label']['position'] = 'right'

    common_options['xAxis'][0]['type'] = 'value'
    del common_options['xAxis'][0]['data']
    common_options['yAxis'][0].update({
        'type': 'category',
        'data': '#set_data#',
    })

    await self._create_trend_chart(
        echart_options=common_options if not clean else horizontal_clean_variant(common_options),
        variant=clean, series_options=series_options, option_modifications=option_modifications,
        axes=x, values=y, data_mapping_to_tuples=data_mapping_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )



from shimoku.actions.execute_action import execute_action
from shimoku.exceptions import ActionError
import unittest


class TestActions(unittest.TestCase):

    def test_action(self):
        action_code = """
from shimoku import Client
import shimoku

class UsesClient:

    def __init__(self):
        print('This should work')

    def use_s(self, shimoku_client: Client):
        shimoku_client.set_workspace()
        shimoku_client.set_menu_path('Pyodide test')
        shimoku_client.plt.gauge_indicator(
            order=2,
            value=len(shimoku_client.menu_paths.get_menu_path_components(name='test-free-echarts')),
            description='Trying to get the number of components',
            title='Pyodide test',
        )

def deep_no_use_s(shimoku_client: Client):
    pass

def no_use_s(shimoku_client: Client):
    deep_no_use_s(shimoku_client)

def deep_deep_use_s():
    shimoku_client.set_workspace()
    shimoku_client.set_menu_path('Pyodide test')
    shimoku_client.plt.gauge_indicator(
        order=2,
        value=len(shimoku_client.menu_paths.get_menu_path_components(name='test-free-echarts')),
        description='Trying to get the number of components',
        title='Pyodide test',
    )
         
def use_s(shimoku_client: Client):
    def deep_use_s():
        deep_deep_use_s()
    deep_use_s()

def action(shimoku_client: Client):
    shimoku_client._async_pool.ACTIONS_TEST = True
    print('This should work')
    no_use_s(shimoku_client)
    use_s(shimoku_client)

    uses_shimoku_client = UsesClient()
    uses_shimoku_client.use_s(shimoku_client)

    shimoku_client.run()
        """

        execute_action(action_code)

        codes_with_errors = [
            """
from shimoku import Client 

# error, needs to be an 'action' function   
""",
            """
# error, needs to import Client
def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client as shimo  # error

def action(shimoku_client: shimo):
    pass
""",
            """
from shimoku import Client
import asyncio  # error

def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client
from asyncio import run  # error

def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client

shimo = Client()  # error

def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client

shimoku_client = 6  # error

def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client

class UsesClient:
    def action(shimoku_client: Client):  # error
        pass
""",
            """
from shimoku import Client

def non_action():
    def action(shimoku_client: Client):  # error
        pass
""",
            """
from shimoku import Client

def action(shimoku_client: Client):
    pass

def action(shimoku_client: Client):  # error
    pass
""",
            """
from shimoku import Client

def action(shimoku_client: Client, other_param: int):  # error
    pass
""",
            """
from shimoku import Client

def action():
    pass
""",
            """
from shimoku import Client

def action(shimo: Client):  # error
    pass
""",
            """
from shimoku import Client

def action(shimoku_client: int):  # error
    pass
""",
            """
from shimoku import Client

def action(shimoku_client):
    pass    
""",
            """
from shimoku import Client

def use_s(non_annotated_param): # error
    pass

def action(shimoku_client: Client):
    use_s(shimoku_client)
""",
            """
from shimoku import Client

def use_s(shimo: Client): # error
    pass
""",
            """
from shimoku import Client

def use_s(shimoku_client: Optional[Client] = None): # error
    pass

def action(shimoku_client: Client):
    use_s()
""",
            """
from shimoku import Client

def use_s(shimoku_client: int):
    pass
""",
            """
from shimoku import Client

async def use_s(shimoku_client: Client):  # error
    pass

def action(shimoku_client: Client):
    use_s(shimoku_client)
""",
            """
from shimoku import Client

def use_s(shimoku_client: Client):
    return shimoku_client  # error

def action(shimoku_client: Client):
    use_s(shimoku_client)
""",
            """
from shimoku import Client

def use_s(other_param: int):
    print(other_param)

def action(shimoku_client: Client):
    use_s(shimoku_client)  # error
""",
            """
from shimoku import Client

def use_s(other_param: int):
    pass

def action(shimoku_client: Client):
    use_s(other_param=shimoku_client)  # error
""",
            """
from shimoku import Client

class UsesClient:

    s = Client()  # error
    
    def use_s(self):
        pass

def action(shimoku_client: Client):
    pass
""",
            """
from shimoku import Client

class UsesClient:
    
    def assign_s(shimoku_client: Client):
        clients_dict = {}
        clients_dict['s'] = shimoku_client  # error

def action(shimoku_client: Client):
    uses_shimoku_client = UsesClient()
    uses_shimoku_client.assign_s(shimoku_client)
    
""",
            ]
        for code_with_errors in codes_with_errors:
            print()
            with self.assertRaises(ActionError):
                execute_action(code_with_errors, True)



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


def code_gen_from_html_string(html_string: str):
    """ Generate code for an html string.
    :param html_string: html string to generate code from
    :return: list of code lines
    """
    code_lines = []
    current_line = ""
    for c in html_string:
        if c in ['\n', '\r']:
            code_lines.append(current_line)
            current_line = ""
        elif c == '<':
            code_lines.append(current_line)
            current_line = '<'
        elif c == ';' and len(current_line) > 60:
            code_lines.append(current_line + ';')
            current_line = ""
        else:
            current_line += c
    if current_line:
        code_lines.append(current_line)

    return [f'"{line}"' for line in code_lines if line]


async def code_gen_from_html(
        self: 'AppCodeGen', report: 'Report'
) -> list[str]:
    """ Generate code for an html report.
    :param report: report to generate code from
    :return: list of code lines
    """

    html = report['chartData'][0]["value"].replace("'", "\\'").replace('"', '\\"')
    html_lines = ['    ' + line for line in code_gen_from_html_string(html)]
    if not html_lines:
        return ['pass']
    html_lines[-1] += ','
    code_lines = [
        'shimoku_client.plt.html(',
        f'    order={report["order"]},',
    ]
    if report['sizeColumns'] != 12:
        code_lines.append(f'    cols_size={report["sizeColumns"]},')
    if report['sizeRows']:
        code_lines.append(f'    rows_size={report["sizeRows"]},')
    if report['sizePadding'] != '0,0,0,0':
        code_lines.append(f'    padding="{report["sizePadding"]}",')

    code_lines.extend([f'    html={html_lines[0][4:]}', *html_lines[1:], ')'])

    return code_lines



from typing import Optional, Dict, TYPE_CHECKING

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.universe import Universe

import logging
logger = logging.getLogger(__name__)


class ActivityTemplate(Resource):

    _module_logger = logger
    resource_type = 'activityTemplate'
    alias_field = ('name', 'version')
    plural = 'activityTemplates'

    def __init__(self, parent: 'Universe', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[Dict] = None):

        params = dict(
            name=alias if alias else '',
            version='',
            description='',
            availableModeCost={},
            inputSettings={},
            minRunInterval=0,
            enabled=False,
            provider='',
            tags=[],
        )

        super().__init__(
            parent=parent, uuid=uuid, db_resource=db_resource,
            params_to_serialize=['availableModeCost', 'inputSettings'], params=params)



from copy import copy
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report

from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_indicators import code_gen_from_indicator
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_echarts import code_gen_from_echarts
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_annotated_echart import code_gen_from_annotated_echart
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_table import code_gen_from_table
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_form import code_gen_from_form
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_html import code_gen_from_html
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_button import code_gen_from_button
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_filter import code_gen_from_filter
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_iframe import code_gen_from_iframe


def delete_default_properties(properties: dict, default_properties: dict) -> dict:
    """ Delete default properties from a report.
    :param properties: properties of a report
    :param default_properties: default properties of a report
    :return: properties without default properties
    """
    properties = copy(properties)
    for key, value in default_properties.items():
        if key not in properties:
            continue
        if properties[key] == value:
            del properties[key]
        if isinstance(value, dict):
            if key not in properties:
                continue
            properties[key] = delete_default_properties(properties[key], value)
            if len(properties[key]) == 0:
                del properties[key]
    return properties


async def code_gen_from_other_reports(
        self: 'AppCodeGen', report: 'Report'
) -> list[str]:
    """ Generate code for a report that is not a tabs group.
    :param report: report to generate code from
    :param is_last: whether the report is the last one
    :return: list of code lines """
    code_lines = []

    properties = delete_default_properties(report['properties'], report.default_properties)
    del properties['hash']

    if report['reportType'] == 'INDICATOR':
        code_lines.extend(await code_gen_from_indicator(self, report, properties))
    elif report['reportType'] == 'ECHARTS2':
        code_lines.extend(await code_gen_from_echarts(self, report, properties))
    elif report['reportType'] == 'TABLE':
        code_lines.extend(await code_gen_from_table(self, report, properties))
    elif report['reportType'] == 'FORM':
        code_lines.extend(await code_gen_from_form(self, report, properties))
    elif report['reportType'] == 'HTML':
        code_lines.extend(await code_gen_from_html(self, report))
    elif report['reportType'] == 'IFRAME':
        code_lines.extend(await code_gen_from_iframe(self, report))
    elif report['reportType'] == 'ANNOTATED_ECHART':
        code_lines.extend(await code_gen_from_annotated_echart(self, report, properties))
    elif report['reportType'] == 'BUTTON':
        code_lines.extend(await code_gen_from_button(self, report))
    elif report['reportType'] == 'FILTERDATASET':
        code_lines.extend(await code_gen_from_filter(self, report))
    else:
        code_lines.extend(
            [f"shimoku_client.add_report({report['reportType']}, order={report['order']}, data=dict())"])

    return code_lines



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_dict
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


async def code_gen_from_indicator(
        self: 'AppCodeGen', report: 'Report', properties: dict
) -> list[str]:
    """ Generate code for an indicator report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    report_dict = report.cascade_to_dict()
    if self._actual_bentobox:
        report_dict['sizeColumns'] += 1
    report_params = self.code_gen_report_params(report_dict)
    properties_code_lines = code_gen_from_dict(properties, 4)
    return [
        'shimoku_client.plt.indicator(',
        *report_params,
        f'    data={properties_code_lines[0][4:]}',
        *properties_code_lines[1:],
        ')'
    ]



from typing import Optional
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from shimoku.cli.utils import choose_from_menu
from shimoku.ai.ai_layer import get_activity_template, ActivityTemplate, AILayer
from pydoc import locate
import logging
logger = logging.getLogger(__name__)


def add_execute_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the get parser to a parser
    :param parser: Parser to add the get parser to
    :return: Get parser
    """
    params = {
        'name': 'execute',
        'description': 'Commands to execute activities',
    }
    if parser:
        execute_parser = parser.add_command(**params)
    else:
        execute_parser = CLIParser(**params)
    common_args = [
        CLIFuncParam(name='local-port', arg_type=int, arg_help='Local port to use', mandatory=False),
        CLIFuncParam(name='environment', arg_type=str, arg_help='Environment to use', mandatory=False),
        CLIFuncParam(name='access-token', arg_type=str, arg_help='Access token to use', mandatory=False),
        CLIFuncParam(name='universe-id', arg_type=str, arg_help='Universe ID to use', mandatory=False),
    ]

    module_functions = [
        ai_function
    ]

    for func in module_functions:
        execute_parser.decor_add_command(common_args=common_args)(func)

    return execute_parser


async def ai_function(
    universe_api_key: Optional[str],
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    ai_function: str = CLIFuncParam(prompt=True),
    version: str = CLIFuncParam(default=None, mandatory=False),
    **kwargs
):
    """ Execute any ai function
    :param universe_api_key: The universe api key that the ai funcion will use
    :param workspace_id: The workspace id
    :param menu_path: The id or name of the menu path
    :param ai_function: The name of the ai function
    :param version: The version of the ai function
    """
    async def get_params(
            _ai_layer: AILayer,
            _activity_template: ActivityTemplate
    ):
        """ Get the parameters interactively
        :param _ai_layer: AILayer
        :param _activity_template: ActivityTemplate
        """
        params = {}
        input_settings = _activity_template['inputSettings']
        for param_name, definition in input_settings.items():
            param_value = None
            while param_value is None:
                input_str = (f"Enter value for {param_name}\n"
                             f"  type: {definition['datatype']}\n"
                             f"  description: {definition['description']})\n")
                if not definition['mandatory']:
                    input_str = input_str[:-1] + "Press enter to skip\n"
                if definition['datatype'] == 'file':
                    input_file_names = [file['file_name']
                                        for file in await _ai_layer.get_available_input_files()] + ['']
                    param_value = choose_from_menu(input_file_names, input_str)
                elif definition['datatype'] == 'model':
                    model_names = [model['file_name'] for model in await _ai_layer.get_available_models()] + ['']
                    param_value = choose_from_menu(model_names, input_str)
                else:
                    param_value = input(input_str)
                    if param_value and definition['datatype'] != 'str':
                        try:
                            param_value = eval(f"{definition['datatype']}({param_value})")
                        except ValueError:
                            logger.error(f"Invalid value for {param_name}")
                            param_value = None

                if not param_value:
                    if definition['mandatory']:
                        param_value = None
                        print(f"{param_name} is mandatory")
                    else:
                        continue

                params[param_name] = param_value

        return params

    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    universe = await resource_getter.get_universe()
    activity_template = await get_activity_template(universe, ai_function, version)
    ai_layer = await resource_getter.get_ai_layer()
    await ai_layer.generic_execute(
        ai_function=ai_function,
        version=version,
        universe_api_key=universe_api_key,
        **(await get_params(ai_layer, activity_template))
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(add_execute_parser().parse_args())



import datetime as dt
import pandas as pd
import asyncio
from copy import deepcopy
from typing import Optional, Union, TYPE_CHECKING, TypeVar
from numbers import Number


from ..base_resource import Resource
from shimoku.exceptions import DataError

if TYPE_CHECKING:
    from .app import App

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)

Mapping = TypeVar('Mapping', bound=Union[dict, list[str], str])


def get_column_types(data_point: dict, sort: Optional[dict]) -> dict:
    """ Given a data point, return a dictionary with the column names as keys and their types as values
    :param data_point: a data point
    :param sort: a dictionary with the sort field and order
    """
    data_mapping = {}
    str_counter = 0
    float_counter = 0
    date_counter = 0
    for k, v in data_point.items():
        if str_counter > 50:
            raise ValueError('Too many string fields')
        if float_counter > 50:
            raise ValueError('Too many int fields')
        if date_counter > 50:
            raise ValueError('Too many date fields')
        if isinstance(v, str) or isinstance(v, bool):
            str_counter += 1
            data_mapping.update({k: f'stringField{str_counter}'})
        elif isinstance(v, Number):
            if sort and k == sort['field']:
                data_mapping.update({k: 'orderField1'})
                sort['field'] = 'orderField1'
            else:
                float_counter += 1
                data_mapping.update({k: f'intField{float_counter}'})
        elif isinstance(v, (dt.date, dt.datetime, pd.Timestamp)):
            date_counter += 1
            data_mapping.update({k: f'dateField{date_counter}'})
            if sort and k == sort['field']:
                sort['field'] = f'dateField{date_counter}'
        elif isinstance(v, dict):
            data_mapping.update({k: f'customField1'})
        else:
            log_error(logger, f'Unknown value type {v} | Type {type(v)} | Key {k}', DataError)
    return data_mapping


def change_keys_df(df: pd.DataFrame, mapping: dict) -> list[dict]:
    """ Given a data and a mapping, change the keys of the data to the values of the mapping
    :param df: data to change the keys
    :param mapping: mapping of the keys to change
    """
    df = df.rename(columns=mapping)
    for k in mapping.values():
        # test if all the values are of the same type
        if 'int' in k or 'order' in k:
            try:
                df[k] = df[k].astype(float)
            except ValueError:
                log_error(logger, f'Column {k} contains values that cannot be converted to float', DataError)
        elif 'string' in k:
            try:
                df[k] = df[k].astype(str)
            except ValueError:
                log_error(logger, f'Column {k} contains values that cannot be converted to string', DataError)
        elif 'date' in k:
            try:
                df[k] = pd.to_datetime(df[k]).dt.strftime('%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                log_error(logger,f'Column {k} contains values that cannot be converted to date ', DataError)

    return df.to_dict('records')


def convert_input_data_to_db_items(
    data: Union[pd.DataFrame, dict],
    sort: Optional[dict] = None, dump_whole: bool = False,
    column_types: Optional[dict] = None
) -> Union[list[dict], dict]:
    """Given an input data, for all the keys of the data convert it to
     a Shimoku body parameter for data table

      stringField1: String
      stringField2: String
      stringField3: String
      stringField4: String
      intField1: Float
      intField2: Float
      intField3: Float
      intField4: Float
      dateField1: AWSDateTime
      customField1: AWSJSON

    Example
    ------------
    input
        data = [
         {'product': 'Matcha Latte', '2015': 43.3, '2016': 85.8, '2017': 93.7},
         {'product': 'Milk Tea', '2015': 83.1, '2016': 73.4, '2017': 55.1},
         {'product': 'Cheese Cocoa', '2015': 86.4, '2016': 65.2, '2017': 82.5},
         {'product': 'Walnut Brownie', '2015': 72.4, '2016': 53.9, '2017': 39.1}
        ]

    intermediate output
        d = {
            'product': 'stringField1',
            '2015': 'intField1',
            '2016': 'intField2',
            '2017': 'intField3'
        }

    output
        [
            {'stringField1': 'Matcha Latte',
             'intField1': 43.3,
             'intField2': 85.8,
             'intField3': 93.7
             },
            {'stringField1': 'Milk Tea',
             'intField1': 83.1,
             'intField2': 73.4,
             'intField3': 55.1
             },
            {'stringField1': 'Cheese Cocoa',
             'intField1': 86.4,
             'intField2': 65.2,
             'intField3': 82.5
             },
            {'stringField1': 'Walnut Brownie',
             'intField1': 72.4,
             'intField2': 53.9,
             'intField3': 39.1
             }
        ]
    """
    if dump_whole:
        return {'customField1': data}
    elif isinstance(data, pd.DataFrame):
        first_element = data.iloc[0].to_dict()
        column_mapping = get_column_types(first_element, sort) if column_types is None else column_types
        if data.isnull().values.any():
            log_error(logger, 'Data contains null values please check for missing values', DataError)
        return change_keys_df(data, column_mapping)
    else:
        log_error(logger, f'Unknown data type {type(data)}', DataError)


class DataSet(Resource):
    """
    DataSet resource class
    """

    _module_logger = logger
    resource_type = 'dataSet'
    alias_field = 'name'
    plural = 'dataSets'

    class DataPoint(Resource):

        _module_logger = logger
        resource_type = 'data'
        plural = 'datas'

        def __init__(self, parent: 'DataSet', uuid: Optional[str] = None, db_resource: Optional[dict] = None):

            params = dict(
                dataSetId=parent['id'],
                orderField1=None,
                description=None,
                customField1=None,
            )
            for i in range(1, 6):
                params.update({f'dateField{i}': None})
            for i in range(1, 51):
                params.update({f'stringField{i}': None})
                params.update({f'intField{i}': None})

            super().__init__(parent=parent, db_resource=db_resource, uuid=uuid, params=params,
                             params_to_serialize=['customField1'])

        async def delete(self):
            """ Delete the data point """
            return await self._base_resource.delete()

        async def update(self):
            """ Update the data point """
            return await self._base_resource.update()

    def __init__(self, parent: 'App', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[dict] = None):

        params = dict(
            name=alias,
            columns=None,
        )
        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, children=[self.DataPoint],
                         check_params_before_creation=['name'], params=params,
                         params_to_serialize=['columns'])

    def delete(self):
        """ Delete the data point """
        return self._base_resource.delete()

    async def update(self):
        """ Update the data point """
        return await self._base_resource.update()

    # DataPoint methods
    async def create_data_points(
            self, data_points: Union[pd.DataFrame, dict], sort: Optional[dict] = None,
            dump_whole: bool = False, column_types: Optional[dict] = None
    ) -> tuple[Mapping, dict]:
        """ Create data points for a report
        :param data_points: data points to be created
        :param sort: sort parameter
        :param dump_whole: whether to dump the whole data into a single field
        :param column_types: precomputed column types
        :return: mapping
        """
        copy_sort = deepcopy(sort) if sort else None
        converted_data_points = convert_input_data_to_db_items(
            data_points, copy_sort, dump_whole=dump_whole, column_types=column_types
        )
        if isinstance(converted_data_points, dict):
            converted_data_points = [converted_data_points]
        await self._base_resource.create_children_batch(self.DataPoint, converted_data_points, unit=' data points')
        keys = [k for k in converted_data_points[0].keys() if not copy_sort or copy_sort['field'] != k]
        if self.api_client.playground and isinstance(data_points, (pd.DataFrame, list)) and len(data_points) > 0:
            if isinstance(data_points, pd.DataFrame):
                data_points = data_points.to_dict('records')
            self['columns'] = {k: m for k, m in zip(data_points[0].keys(), converted_data_points[0].keys())}
            await self.update()
        return keys, copy_sort

    async def delete_data_points(self):
        """ Delete data points """
        self.clear()
        data_points = await self._base_resource.get_children(self.DataPoint)
        if data_points:
            await asyncio.gather(*[self._base_resource.delete_child(self.DataPoint, uuid=dp['id'])
                                   for dp in data_points])

    async def get_data_points(self, limit: Optional[int] = None):
        """ Get data points """
        self.clear()
        return await self._base_resource.get_children(self.DataPoint, limit=limit)

    async def get_one_data_point(self) -> Optional['DataSet.DataPoint']:
        """ Get one data point """
        self.clear()
        data_points = await self._base_resource.get_children(self.DataPoint, limit=1)
        return data_points[0] if data_points else None



from shimoku.utils import create_function_name
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_other import delete_default_properties
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_tabs import code_gen_tabs_and_other
from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_value
from shimoku.api.resources.reports.modal import Modal
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen


async def code_gen_from_modal(self: 'AppCodeGen', tree: dict) -> list[str]:
    """ Generate code for a modal.
    :param tree: tree of reports
    :return: list of code lines
    """
    # code_lines = (await code_gen_tabs_functions(tree['tab_groups']))
    code_lines = []
    modal: Modal = tree['modal']
    properties = delete_default_properties(modal['properties'], Modal.default_properties)
    properties['modal_name'] = tree['name']
    del properties['hash']
    if 'reportIds' in properties:
        del properties['reportIds']
    if 'open' in properties:
        if properties['open']:
            properties['open_by_default'] = True
        del properties['open']
    code_lines.extend([
        'shimoku_client.plt.set_modal(',
        *[f'    {k}={code_gen_from_value(v)},' for k, v in properties.items()],
        ')',
    ])
    code_lines.extend(await code_gen_tabs_and_other(self, tree))
    code_lines.extend(['', 'shimoku_client.plt.pop_out_of_modal()'])
    return code_lines


async def code_gen_modals_functions(self: 'AppCodeGen', path: str) -> list[str]:
    """ Generate code for modals.
    :param path: path to the modals
    :return: list of code lines
    """
    code_lines = []
    for modal in self._code_gen_tree.tree[path]['modals']:
        modal_code_lines = await code_gen_from_modal(self, modal)
        code_lines.extend([
            '',
            f'def modal_{create_function_name(modal["name"])}'
            f'(shimoku_client: Client):',
            *['    ' + line for line in modal_code_lines]
        ])
    return code_lines



from typing import Optional, TYPE_CHECKING

import pandas as pd

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def speed_gauge_chart(
    self: 'PlotLayer', order: int, name: str, value: int,
    min_value: int, max_value: int, title: Optional[str] = None, rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3, padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a speed gauge chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['yAxis']
    del common_options['xAxis']
    del common_options['grid']
    del common_options['toolbox']['feature']['magicType']
    common_options['grid'] = {'top': '1%', 'left': '1%', 'right': '1%'}
    common_options['tooltip']['trigger'] = 'item'
    series_options.update({
        'type': 'gauge',
        'min': min_value,
        'max': max_value,
        'progress': {'show': True, 'width': 18},
        'axisLine': {'lineStyle': {'width': 18}},
        'axisTick': {'show': False},
        'splitLine': {'length': 15, 'lineStyle': {'width': 2, 'color': '#999'}},
        'axisLabel': {'distance': 25, 'color': '#999', 'fontSize': 10},
        'anchor': {'show': True, 'showAbove': True, 'size': 25, 'itemStyle': {'borderWidth': 10}},
        'title': {'show': True, 'offsetCenter': [0, '45%'], 'fontSize': 20},
        'detail': {'valueAnimation': True, 'fontSize': 20, 'offsetCenter': [0, '70%']},
    })
    common_options['series'] = [series_options]

    await self._create_echart(
        fields=[{'name': 'name', 'value': 'value'}], options=common_options, option_modifications=option_modifications,
        data_mapping_to_tuples=await self._choose_data(order, pd.DataFrame([{'name': name, 'value': value}])),
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding)



from __future__ import annotations

import fnmatch
import re
import dateutil.parser
from typing import List, Union, Type, Callable
import graphene
from dataclasses import field
from functools import partial


class AccountExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: AccountExposed)


class ReportExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: ReportExposed)


class FileExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: FileExposed)


class ActivityExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: ActivityExposed)


class DataSetExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: DataSetExposed)


class RolePermissionExposedList(graphene.ObjectType):
    items = graphene.List(lambda: RolePermissionExposed)


class AppDashboardExposedList(graphene.ObjectType):
    items = graphene.List(lambda: AppDashboardExposed)


class AppTypeExposed(graphene.ObjectType):
    id = graphene.String()
    key = graphene.String()
    name = graphene.String()
    appTypeUniverseId = graphene.String()
    normalizedName = graphene.String()


class AppExposed(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    trialDays = graphene.Int()
    isDisabled = graphene.Boolean()
    appBusinessId = graphene.String()
    createdAt = graphene.String()
    churnedDate = graphene.String()
    reports = graphene.Field(lambda: ReportExposedList)
    hideTitle = graphene.Boolean()
    paymentType = graphene.String()
    order = graphene.Int()
    normalizedName = graphene.String()
    files = graphene.Field(lambda: FileExposedList)
    hidePath = graphene.Boolean()
    showBreadcrumb = graphene.Boolean()
    showHistoryNavigation = graphene.Boolean()
    activities = graphene.Field(lambda: ActivityExposedList)
    dataSets = graphene.Field(lambda: DataSetExposedList)
    rolePermissions = graphene.Field(lambda: RolePermissionExposedList)
    appDashboards = graphene.Field(lambda: AppDashboardExposedList)
    universeId = graphene.String()
    type = graphene.Field(lambda: AppTypeExposed)


class DashboardExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: DashboardExposed)


class BusinessExposed(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    type = graphene.String()
    createdAt = graphene.String()
    etlMachineLastProcessed = graphene.String()
    integrationId = graphene.String()
    defaultAppPaymentType = graphene.String()
    defaultstr = graphene.String()
    theme = graphene.String()
    roles = graphene.List(lambda: RolePermissionExposed)
    rolePermissions = graphene.Field(lambda: RolePermissionExposedList)
    businessUniverseId = graphene.String()
    dashboards = graphene.Field(lambda: DashboardExposedList)
    apps = graphene.Field(lambda: AppExposedList)
    modules = graphene.Field(lambda: ModuleExposedList)
    universe = graphene.Field(lambda: UniverseFilteredExposed)


class ModuleExposed(graphene.ObjectType):
    enabled = graphene.Boolean()
    id = graphene.String()
    type = graphene.String()


class ModuleExposedList(graphene.ObjectType):
    items = graphene.List(lambda: ModuleExposed)


class AccountExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    startupAppId = graphene.String()
    userType = graphene.String()
    notificationsLastRead = graphene.String()
    startupApp = graphene.Field(lambda: AppExposed)
    business = graphene.Field(BusinessExposed)
    shopifyToken = graphene.String()
    roles = graphene.List(graphene.String)


class ReportDataSetExposedList(graphene.ObjectType):
    items = graphene.List(lambda: ReportDataSetExposed)


class ReportExposed(graphene.ObjectType):
    id = graphene.String()
    appId = graphene.String()
    app = graphene.Field(lambda: AppExposed)
    createdAt = graphene.String()
    order = graphene.Int()
    grid = graphene.String()
    sizeRows = graphene.Int()
    sizeColumns = graphene.Int()
    sizePadding = graphene.String()
    isDisabled = graphene.Boolean()
    reportType = graphene.String()
    path = graphene.String()
    pathOrder = graphene.Int()
    dataFields = graphene.String()
    title = graphene.String()
    description = graphene.String()
    chartData = graphene.String()
    chartDataItem = graphene.String()
    chartDataAux = graphene.String()
    smartFilters = graphene.String()
    subscribe = graphene.Boolean()
    reportDataSets = graphene.Field(lambda: ReportDataSetExposedList)
    properties = graphene.String()
    bentobox = graphene.String()
    hidePath = graphene.Boolean()
    showBreadcrumb = graphene.Boolean()
    showHistoryNavigation = graphene.Boolean()
    universeId = graphene.String()

    @staticmethod
    def list_resolver_extra(
            publicPermissions: DashboardPublicPermissionInputExposed = None,
            **kwargs
    ):
        items = kwargs.get('items')
        return items


class DataExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: DataExposed)


class DataSetExposed(graphene.ObjectType):
    id = graphene.String()
    data = graphene.Field(lambda: DataExposedList)
    reportDataSetId = graphene.String()
    reportDataSets = graphene.Field(lambda: ReportDataSetExposedList)
    dataSetAppId = graphene.String()
    app = graphene.Field(lambda: AppExposed)
    universeId = graphene.String()
    name = graphene.String()
    columns = graphene.String()


class ReportDataSetExposed(graphene.ObjectType):
    id = graphene.String()
    dataSetId = graphene.String()
    dataSet = graphene.Field(lambda: DataSetExposed)
    reportId = graphene.String()
    report = graphene.Field(lambda: ReportExposed)
    properties = graphene.String()


class DashboardPublicPermissionInputExposed(graphene.InputObjectType):
    resourceId = graphene.String()
    token = graphene.String()


class SearchableDataSortableFields(graphene.Enum):
    id = "id"
    createdAt = "createdAt"
    dataSetId = "dataSetId"
    description = "description"
    stringField1 = "stringField1"
    stringField2 = "stringField2"
    stringField3 = "stringField3"
    stringField4 = "stringField4"
    stringField5 = "stringField5"
    stringField6 = "stringField6"
    stringField7 = "stringField7"
    stringField8 = "stringField8"
    stringField9 = "stringField9"
    stringField10 = "stringField10"
    stringField11 = "stringField11"
    stringField12 = "stringField12"
    stringField13 = "stringField13"
    stringField14 = "stringField14"
    stringField15 = "stringField15"
    stringField16 = "stringField16"
    stringField17 = "stringField17"
    stringField18 = "stringField18"
    stringField19 = "stringField19"
    stringField20 = "stringField20"
    stringField21 = "stringField21"
    stringField22 = "stringField22"
    stringField23 = "stringField23"
    stringField24 = "stringField24"
    stringField25 = "stringField25"
    stringField26 = "stringField26"
    stringField27 = "stringField27"
    stringField28 = "stringField28"
    stringField29 = "stringField29"
    stringField30 = "stringField30"
    stringField31 = "stringField31"
    stringField32 = "stringField32"
    stringField33 = "stringField33"
    stringField34 = "stringField34"
    stringField35 = "stringField35"
    stringField36 = "stringField36"
    stringField37 = "stringField37"
    stringField38 = "stringField38"
    stringField39 = "stringField39"
    stringField40 = "stringField40"
    stringField41 = "stringField41"
    stringField42 = "stringField42"
    stringField43 = "stringField43"
    stringField44 = "stringField44"
    stringField45 = "stringField45"
    stringField46 = "stringField46"
    stringField47 = "stringField47"
    stringField48 = "stringField48"
    stringField49 = "stringField49"
    stringField50 = "stringField50"
    intField1 = "intField1"
    intField2 = "intField2"
    intField3 = "intField3"
    intField4 = "intField4"
    intField5 = "intField5"
    intField6 = "intField6"
    intField7 = "intField7"
    intField8 = "intField8"
    intField9 = "intField9"
    intField10 = "intField10"
    intField11 = "intField11"
    intField12 = "intField12"
    intField13 = "intField13"
    intField14 = "intField14"
    intField15 = "intField15"
    intField16 = "intField16"
    intField17 = "intField17"
    intField18 = "intField18"
    intField19 = "intField19"
    intField20 = "intField20"
    intField21 = "intField21"
    intField22 = "intField22"
    intField23 = "intField23"
    intField24 = "intField24"
    intField25 = "intField25"
    intField26 = "intField26"
    intField27 = "intField27"
    intField28 = "intField28"
    intField29 = "intField29"
    intField30 = "intField30"
    intField31 = "intField31"
    intField32 = "intField32"
    intField33 = "intField33"
    intField34 = "intField34"
    intField35 = "intField35"
    intField36 = "intField36"
    intField37 = "intField37"
    intField38 = "intField38"
    intField39 = "intField39"
    intField40 = "intField40"
    intField41 = "intField41"
    intField42 = "intField42"
    intField43 = "intField43"
    intField44 = "intField44"
    intField45 = "intField45"
    intField46 = "intField46"
    intField47 = "intField47"
    intField48 = "intField48"
    intField49 = "intField49"
    intField50 = "intField50"
    dateField1 = "dateField1"
    dateField2 = "dateField2"
    dateField3 = "dateField3"
    dateField4 = "dateField4"
    dateField5 = "dateField5"
    dateField6 = "dateField6"
    dateField7 = "dateField7"
    dateField8 = "dateField8"
    customField1 = "customField1"
    orderField1 = "orderField1"


class SearchableSortDirection(graphene.Enum):
    asc = "asc"
    desc = "desc"


class SearchableDataSortInput(graphene.InputObjectType):
    field = graphene.Field(lambda: SearchableDataSortableFields)
    direction = graphene.Field(lambda: SearchableSortDirection)


class SearchableIDFilterInput(graphene.InputObjectType):
    ne = graphene.String()
    gt = graphene.String()
    lt = graphene.String()
    gte = graphene.String()
    lte = graphene.String()
    eq = graphene.String()
    match = graphene.String()
    matchPhrase = graphene.String()
    matchPhrasePrefix = graphene.String()
    multiMatch = graphene.String()
    exists = graphene.Boolean()
    wildcard = graphene.String()
    regexp = graphene.String()
    range = graphene.List(graphene.String)


class StringOrStringList(graphene.Scalar):
    """
    Custom Scalar to handle both String and List of Strings
    """

    @staticmethod
    def serialize(value):
        return value

    @staticmethod
    def parse_value(value):
        if not isinstance(value, list):
            return [value]  # Convert single string to a list
        return value

    @staticmethod
    def parse_literal(ast):
        # Handle parsing GraphQL literals
        if isinstance(ast, graphene.StringValue):
            return [ast.value]
        elif isinstance(ast, graphene.ListValue):
            return [elem.value for elem in ast.values if isinstance(elem, graphene.StringValue)]


class SearchableStringFilterInput(graphene.InputObjectType):
    ne = graphene.Field(StringOrStringList)
    gt = graphene.Field(StringOrStringList)
    lt = graphene.Field(StringOrStringList)
    gte = graphene.Field(StringOrStringList)
    lte = graphene.Field(StringOrStringList)
    eq = graphene.Field(StringOrStringList)
    match = graphene.Field(StringOrStringList)
    matchPhrase = graphene.Field(StringOrStringList)
    matchPhrasePrefix = graphene.Field(StringOrStringList)
    multiMatch = graphene.Field(StringOrStringList)
    exists = graphene.Boolean()
    wildcard = graphene.Field(StringOrStringList)
    regexp = graphene.Field(StringOrStringList)
    range = graphene.Field(StringOrStringList)


class FloatOrString(graphene.Scalar):
    """
    Custom Scalar type to handle both Float and String types
    """

    @staticmethod
    def serialize(value):
        return value

    @staticmethod
    def parse_value(value):
        try:
            return float(value)
        except ValueError:
            return str(value)

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, (graphene.FloatValue, graphene.StringValue)):
            return ast.value


class SearchableFloatFilterInput(graphene.InputObjectType):
    # Front end sends strings
    ne = graphene.Field(FloatOrString)
    gt = graphene.Field(FloatOrString)
    lt = graphene.Field(FloatOrString)
    gte = graphene.Field(FloatOrString)
    lte = graphene.Field(FloatOrString)
    eq = graphene.Field(FloatOrString)
    range = graphene.List(FloatOrString)


class SearchableStringFilterInputOrList(graphene.InputObjectType):
    items = graphene.List(SearchableStringFilterInput)

    @staticmethod
    def parse_value(value):
        if not isinstance(value, list):
            return [value]
        return value


class SearchableDataFilterInput(graphene.InputObjectType):
    id = graphene.Field(lambda: SearchableIDFilterInput)
    createdAt = graphene.Field(lambda: SearchableStringFilterInput)
    dataSetId = graphene.Field(lambda: SearchableIDFilterInput)
    description = graphene.Field(lambda: SearchableStringFilterInput)
    stringField1 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField2 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField3 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField4 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField5 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField6 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField7 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField8 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField9 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField10 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField11 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField12 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField13 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField14 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField15 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField16 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField17 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField18 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField19 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField20 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField21 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField22 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField23 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField24 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField25 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField26 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField27 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField28 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField29 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField30 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField31 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField32 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField33 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField34 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField35 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField36 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField37 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField38 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField39 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField40 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField41 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField42 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField43 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField44 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField45 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField46 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField47 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField48 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField49 = graphene.Field(lambda: SearchableStringFilterInput)
    stringField50 = graphene.Field(lambda: SearchableStringFilterInput)
    intField1 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField2 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField3 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField4 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField5 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField6 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField7 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField8 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField9 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField10 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField11 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField12 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField13 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField14 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField15 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField16 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField17 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField18 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField19 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField20 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField21 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField22 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField23 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField24 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField25 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField26 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField27 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField28 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField29 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField30 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField31 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField32 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField33 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField34 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField35 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField36 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField37 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField38 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField39 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField40 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField41 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField42 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField43 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField44 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField45 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField46 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField47 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField48 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField49 = graphene.Field(lambda: SearchableFloatFilterInput)
    intField50 = graphene.Field(lambda: SearchableFloatFilterInput)
    dateField1 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField2 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField3 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField4 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField5 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField6 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField7 = graphene.Field(lambda: SearchableStringFilterInput)
    dateField8 = graphene.Field(lambda: SearchableStringFilterInput)
    customField1 = graphene.Field(lambda: SearchableStringFilterInput)
    orderField1 = graphene.Field(lambda: SearchableFloatFilterInput)

    @classmethod
    def add_fields(cls) -> Type[SearchableDataFilterInput]:
        """ This is necessary because of the way the filter is defined in the API, these keywords are reserved
        in python so we have to add them programmatically.
        """
        add_fields = [
            ('and', graphene.List(SearchableDataFilterInput)),
            ('or', graphene.List(SearchableDataFilterInput)),
            ('not', graphene.Field(SearchableDataFilterInput))
        ]
        for field_name, field_type in add_fields:
            setattr(cls, field_name, field_type)
            if isinstance(field_type, graphene.List):
                cls._meta.fields[field_name] = graphene.InputField(field_type)
            else:
                cls._meta.fields[field_name] = field_type

        return cls


class DataExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    dataSetId = graphene.String()
    dataSet = graphene.Field(lambda: DataSetExposed)
    description = graphene.String()
    stringField1 = graphene.String()
    stringField2 = graphene.String()
    stringField3 = graphene.String()
    stringField4 = graphene.String()
    stringField5 = graphene.String()
    stringField6 = graphene.String()
    stringField7 = graphene.String()
    stringField8 = graphene.String()
    stringField9 = graphene.String()
    stringField10 = graphene.String()
    stringField11 = graphene.String()
    stringField12 = graphene.String()
    stringField13 = graphene.String()
    stringField14 = graphene.String()
    stringField15 = graphene.String()
    stringField16 = graphene.String()
    stringField17 = graphene.String()
    stringField18 = graphene.String()
    stringField19 = graphene.String()
    stringField20 = graphene.String()
    stringField21 = graphene.String()
    stringField22 = graphene.String()
    stringField23 = graphene.String()
    stringField24 = graphene.String()
    stringField25 = graphene.String()
    stringField26 = graphene.String()
    stringField27 = graphene.String()
    stringField28 = graphene.String()
    stringField29 = graphene.String()
    stringField30 = graphene.String()
    stringField31 = graphene.String()
    stringField32 = graphene.String()
    stringField33 = graphene.String()
    stringField34 = graphene.String()
    stringField35 = graphene.String()
    stringField36 = graphene.String()
    stringField37 = graphene.String()
    stringField38 = graphene.String()
    stringField39 = graphene.String()
    stringField40 = graphene.String()
    stringField41 = graphene.String()
    stringField42 = graphene.String()
    stringField43 = graphene.String()
    stringField44 = graphene.String()
    stringField45 = graphene.String()
    stringField46 = graphene.String()
    stringField47 = graphene.String()
    stringField48 = graphene.String()
    stringField49 = graphene.String()
    stringField50 = graphene.String()
    intField1 = graphene.Float()
    intField2 = graphene.Float()
    intField3 = graphene.Float()
    intField4 = graphene.Float()
    intField5 = graphene.Float()
    intField6 = graphene.Float()
    intField7 = graphene.Float()
    intField8 = graphene.Float()
    intField9 = graphene.Float()
    intField10 = graphene.Float()
    intField11 = graphene.Float()
    intField12 = graphene.Float()
    intField13 = graphene.Float()
    intField14 = graphene.Float()
    intField15 = graphene.Float()
    intField16 = graphene.Float()
    intField17 = graphene.Float()
    intField18 = graphene.Float()
    intField19 = graphene.Float()
    intField20 = graphene.Float()
    intField21 = graphene.Float()
    intField22 = graphene.Float()
    intField23 = graphene.Float()
    intField24 = graphene.Float()
    intField25 = graphene.Float()
    intField26 = graphene.Float()
    intField27 = graphene.Float()
    intField28 = graphene.Float()
    intField29 = graphene.Float()
    intField30 = graphene.Float()
    intField31 = graphene.Float()
    intField32 = graphene.Float()
    intField33 = graphene.Float()
    intField34 = graphene.Float()
    intField35 = graphene.Float()
    intField36 = graphene.Float()
    intField37 = graphene.Float()
    intField38 = graphene.Float()
    intField39 = graphene.Float()
    intField40 = graphene.Float()
    intField41 = graphene.Float()
    intField42 = graphene.Float()
    intField43 = graphene.Float()
    intField44 = graphene.Float()
    intField45 = graphene.Float()
    intField46 = graphene.Float()
    intField47 = graphene.Float()
    intField48 = graphene.Float()
    intField49 = graphene.Float()
    intField50 = graphene.Float()
    dateField1 = graphene.String()
    dateField2 = graphene.String()
    dateField3 = graphene.String()
    dateField4 = graphene.String()
    dateField5 = graphene.String()
    dateField6 = graphene.String()
    dateField7 = graphene.String()
    dateField8 = graphene.String()
    customField1 = graphene.String()
    orderField1 = graphene.Float()
    universeId = graphene.String()

    @staticmethod
    def filter_func_template(field_value, filter_value, type_converter, comparator):
        if isinstance(filter_value, list):
            return any([comparator(type_converter(field_value), type_converter(value)) for value in filter_value])
        else:
            return comparator(type_converter(field_value), type_converter(filter_value))

    @staticmethod
    def filter_field(
            items: list, field: str,
            filterFieldInput: Union[SearchableIDFilterInput, SearchableStringFilterInput, SearchableFloatFilterInput]
    ) -> list:
        type_converter: Callable = lambda x: x
        if 'dateField' in field:
            # Set the values lower than an hour to 0
            type_converter: Callable = lambda x: dateutil.parser.parse(x).replace(
                minute=0, second=0, microsecond=0, tzinfo=dateutil.tz.tzutc()
            )
        elif 'intField' in field or 'orderField' in field:
            type_converter: Callable = lambda x: float(x)

        filters: List[Callable] = []
        func = DataExposed.filter_func_template
        if filterFieldInput.ne is not None:
            filters.append(partial(func, filter_value=filterFieldInput.ne,
                                   type_converter=type_converter, comparator=lambda x, y: x != y))
        if filterFieldInput.gt is not None:
            filters.append(partial(func, filter_value=filterFieldInput.gt,
                                   type_converter=type_converter, comparator=lambda x, y: x > y))
        if filterFieldInput.lt is not None:
            filters.append(partial(func, filter_value=filterFieldInput.lt,
                                   type_converter=type_converter, comparator=lambda x, y: x < y))
        if filterFieldInput.gte is not None:
            filters.append(partial(func, filter_value=filterFieldInput.gte,
                                   type_converter=type_converter, comparator=lambda x, y: x >= y))
        if filterFieldInput.lte is not None:
            filters.append(partial(func, filter_value=filterFieldInput.lte,
                                   type_converter=type_converter, comparator=lambda x, y: x <= y))
        if filterFieldInput.eq is not None:
            filters.append(partial(func, filter_value=filterFieldInput.eq,
                                   type_converter=type_converter, comparator=lambda x, y: x == y))
        if isinstance(filterFieldInput, (SearchableIDFilterInput, SearchableStringFilterInput)):
            if filterFieldInput.match is not None:
                filters.append(partial(func, filter_value=filterFieldInput.match,
                                       type_converter=type_converter, comparator=lambda x, y: x == y))
            if filterFieldInput.matchPhrase is not None:
                filters.append(partial(func, filter_value=filterFieldInput.matchPhrase,
                                       type_converter=type_converter, comparator=lambda x, y: x in y))
            if filterFieldInput.matchPhrasePrefix is not None:
                filters.append(partial(func, filter_value=filterFieldInput.matchPhrasePrefix,
                                       type_converter=type_converter, comparator=lambda x, y: x.startswith(y)))
            if filterFieldInput.multiMatch is not None:
                filters.append(partial(func, filter_value=filterFieldInput.multiMatch,
                                       type_converter=type_converter, comparator=lambda x, y: x in y))
            if filterFieldInput.exists:
                filters.append(lambda x: x is not None)
            if filterFieldInput.wildcard:
                filters.append(partial(func, filter_value=filterFieldInput.wildcard,
                                       type_converter=type_converter, comparator=lambda x, y: fnmatch.fnmatch(x, y)))
            if filterFieldInput.regexp:
                filters.append(partial(func, filter_value=filterFieldInput.regexp,
                                       type_converter=type_converter, comparator=lambda x, y: re.match(y, x)))
        return [item for item in items if all([filter_func(getattr(item, field)) for filter_func in filters])]

    @staticmethod
    def filter(input_items, filterInput: SearchableDataFilterInput) -> list:
        not_results = []
        or_results = []
        results = [item for item in input_items]
        filtered = False
        for key, value in filterInput.__dict__.items():
            if value is not None:
                if key == 'and':
                    if len(value) == 0:
                        results = []
                    for _filter in value:
                        results = DataExposed.filter(results, _filter)
                    filtered = True
                elif key == 'or':
                    for _filter in value:
                        or_results.extend(DataExposed.filter(input_items, _filter))
                elif key == 'not':
                    not_results.append(DataExposed.filter(results, value))
                    filtered = True
                else:
                    results = DataExposed.filter_field(results, key, value)
                    filtered = True

        if not filtered:
            results = []
        elif not_results:
            results = [item for item in results if item not in not_results[0]]

        for or_result in or_results:
            if or_result not in results:
                results.append(or_result)

        return results

    @staticmethod
    def list_resolver_extra(
            publicPermissions: DashboardPublicPermissionInputExposed = None,
            sort: SearchableDataSortInput = None,
            filter: SearchableDataFilterInput = None,
            **kwargs
    ):
        items = kwargs.get('items')
        if sort is not None and sort.field is not None and sort.direction is not None:
            sort_direction = sort.direction.value
            sort_field = sort.field.value
            items = sorted(items, key=lambda x: getattr(x, sort_field), reverse=sort_direction == 'desc')
        if filter is not None:
            items = DataExposed.filter(items, filter)
        return items


class FileExposed(graphene.ObjectType):
    id = graphene.String()
    appId = graphene.String()
    app = graphene.Field(AppExposed)
    fileName = graphene.String()
    name = graphene.String()
    createdAt = graphene.String()
    updatedAt = graphene.String()
    universeId = graphene.String()
    contentType = graphene.String()


class RunExposedList(graphene.ObjectType):
    items = graphene.List(lambda: RunExposed)


class ActivityExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    appId = graphene.String()
    app = graphene.Field(lambda: AppExposed)
    runs = graphene.Field(lambda: RunExposedList)
    name = graphene.String()
    settings = graphene.String()
    universeId = graphene.String()
    activityWebHook = graphene.Field(lambda: ActivityWebHookExposed)


class LogExposedList(graphene.ObjectType):
    items = graphene.List(lambda: LogExposed)


class RunExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    activityId = graphene.String()
    activity = graphene.Field(lambda: ActivityExposed)
    settings = graphene.String()
    logs = graphene.Field(lambda: LogExposedList)
    isWebhookCalled = graphene.Boolean()


class LogExposed(graphene.ObjectType):
    id = graphene.String()
    message = graphene.String()
    runId = graphene.String()
    run = graphene.Field(lambda: RunExposed)
    dateTime = graphene.String()
    tags = graphene.String()
    severity = graphene.String()


class RolePermissionExposed(graphene.ObjectType):
    role = graphene.String()
    permission = graphene.String()
    resource = graphene.String()
    target = graphene.String()


class PublicPermission(graphene.ObjectType):
    isPublic = graphene.Boolean()
    permission = graphene.String()
    token = graphene.String()


class DashboardExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    appDashboards = graphene.Field(AppDashboardExposedList)
    businessId = graphene.String()
    rolePermissions = graphene.Field(lambda: RolePermissionExposedList)
    publicPermission = graphene.Field(lambda: PublicPermission)
    name = graphene.String()
    order = graphene.Int()
    isDisabled = graphene.Boolean()


class AppDashboardExposed(graphene.ObjectType):
    id = graphene.String()
    createdAt = graphene.String()
    dashboard = graphene.Field(DashboardExposed)
    app = graphene.Field(AppExposed)
    appId = graphene.String()


class UniverseModuleExposed(graphene.ObjectType):
    id = graphene.String()
    type = graphene.String()
    enabled = graphene.Boolean()


class BusinessAppTypeExposed(graphene.ObjectType):
    id = graphene.String()
    businessId = graphene.String()
    appTypeId = graphene.String()
    business = graphene.Field(lambda: BusinessExposed)


class BusinessExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: BusinessExposed)


class UniverseExposed(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    business = graphene.Field(lambda: BusinessExposedList)
    theme = graphene.String()
    activeUniversePlanId = graphene.String()
    activeUniversePlan = graphene.Field(lambda: ActiveUniversePlan)
    modules = graphene.Field(lambda: ModuleExposedList)


class UniverseFilteredExposed(graphene.ObjectType):
    id = graphene.String()
    name = graphene.String()
    business = graphene.Field(lambda: BusinessExposedList)
    theme = graphene.String()
    activeUniversePlanId = graphene.String()
    activeUniversePlan = graphene.Field(lambda: ActiveUniversePlan)
    modules = graphene.Field(lambda: ModuleExposedList)


class ActiveUniversePlan(graphene.ObjectType):
    id = graphene.String()
    planType = graphene.Field(lambda: PlanType)


class PlanType(graphene.ObjectType):
    id = graphene.String()
    limits = graphene.String()
    price = graphene.Int()
    type = graphene.String()


class AppExposedList(graphene.ObjectType):
    total = graphene.Int()
    nextToken = graphene.String()
    items = graphene.List(lambda: AppExposed)


class ActivityWebHookExposed(graphene.ObjectType):
    id = graphene.String()
    url = graphene.String()
    method = graphene.String()
    headers = graphene.String()
    activityId = graphene.String()
    createdAt = graphene.String()
    updatedAt = graphene.String()
    activity = graphene.Field(lambda: ActivityExposed)



from typing import Optional, Union, TYPE_CHECKING

from shimoku.exceptions import BentoboxError
from shimoku.plt.utils import interpret_color, ShimokuPalette

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


def gauge_indicator(
    self: 'PlotLayer', order: int, value: int,
    title: Optional[str] = "", description: Optional[str] = "",
    cols_size: Optional[int] = 6, rows_size: Optional[int] = 1,
    color: Union[str, int] = 1,
):
    """ Create a gauge indicator """
    bentobox_data = {
        'bentoboxId': f'_{order}',
        'bentoboxOrder': order,
        'bentoboxSizeColumns': cols_size,
        'bentoboxSizeRows': rows_size,
    }

    if self._bentobox_data:
        log_error(logger, 'The gauges group uses a bentobox, so it cannot be used inside another bentobox. Please'
                          ' pop out of the current bentobox before using this function.', BentoboxError)
    self._bentobox_data = bentobox_data

    indicator_data = [
        {
            'description': description,
            'title': title,
            'value': '',
            'color': '',
            'align': 'left'
        }
    ]

    self.indicator(
        data=indicator_data, order=order,
        rows_size=8, cols_size=15, padding='1, 1, 0, 1',
    )

    color = interpret_color(color)

    options = {
        'grid': {
            'left': ['5%'],
            'right': ['5%'],
            'top': ['5%'],
            'bottom': ['5%'],
            'containLabel': True
        },
        'series': [
            {
                'data': '#set_data#',
                'type': 'gauge',
                'startAngle': 360,
                'endAngle': 0,
                'radius': '80%',
                'center': ['50%', '40%'],
                'min': 0,
                'max': 100,
                'pointer': {
                    'show': False,
                },
                'progress': {
                    'show': True,
                    'width': 20,
                    'overlap': False,
                    'roundCap': False,
                    'clip': False,
                    'itemStyle': {
                        'borderWidth': 0,
                        'color': color,
                    }
                },
                'splitLine': {
                    'show': False,
                },
                'axisLine': {'lineStyle': {'width': 20}},
                'axisTick': {
                    'show': False
                },
                'axisLabel': {
                    'show': False,
                },
                'title': {
                    'show': False,
                    'fontSize': 16,
                    'fontFamily': 'Rubik',
                },
                'detail': {
                    'fontSize': 24,
                    'fontFamily': 'Rubik',
                    'font': 'inherit',
                    'color': ShimokuPalette.BLACK.value,
                    'borderColor': 'auto',
                    'borderWidth': 0,
                    'formatter': '{value}%',
                    'valueAnimation': True,
                    'offsetCenter': ['0', '60']
                }
            }
        ]
    }

    self.free_echarts(
        data=[{'value': value}],
        fields=[{'value': 'value'}],
        options=options,
        order=order + 1, rows_size=6, cols_size=7,
        padding='1, 0, 1, 0',
    )

    self.pop_out_of_bentobox()



""""""
from pandas import DataFrame

from typing import Optional, Union

from shimoku.api.utils import validate_data_is_pandarable
from shimoku.api.resources.app import App
from shimoku.api.utils import convert_input_data_to_db_items
from shimoku.exceptions import DataError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging
logger = logging.getLogger(__name__)


class DataSetsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the data set level.
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(self, app: Optional[App]):
        self._app = app

    async def append_to_data_set(
        self, data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """
        Appends data to a dataset, if the dataset does not exist it will be created.
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        :return: dataset id
        """

        df = validate_data_is_pandarable(data)

        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        data_point = await data_set.get_one_data_point()

        if data_point:
            converted_data = convert_input_data_to_db_items(df)
            data_point_keys = [col for col in data_point if data_point[col] is not None]
            db_items_keys = set(converted_data[0].keys())
            if [key for key in db_items_keys if key not in data_point_keys]:
                log_error(logger, f"The input data has different fields than the existing data set.", DataError)

        await self._app.append_data_to_data_set(df, uuid=uuid, name=name)

        return data_set['id']

    async def delete_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Delete a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        await self._app.delete_data_set(uuid=uuid, name=name)

    async def replace_data_from_data_set(
        self, data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """ Replace the data in a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        """
        df = validate_data_is_pandarable(data)

        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        await data_set.delete_data_points()

        await self._app.append_data_to_data_set(df, uuid=uuid, name=name)

    async def get_data_from_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        limit: Optional[int] = None
    ) -> list[dict]:
        """ Get the data in a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param limit: limit of the data points to be returned
        :return: data in the dataset
        """
        data_set = await self._app.get_data_set(uuid=uuid, name=name, create_if_not_exists=False)
        if not data_set:
            logger.warning(f'Data set {name if name else uuid} does not exist.')
            return []
        return [{k: v for k, v in dp.cascade_to_dict().items() if v is not None}
                for dp in await data_set.get_data_points(limit)]



from ..report import Report

import asyncio
import logging
logger = logging.getLogger(__name__)


class Modal(Report):
    """ Modal report class """

    report_type = 'MODAL'

    default_properties = dict(
        **Report.default_properties,
        open=False,
        reportIds=[],
        width=60,
        height=50,
    )

    def __init__(self, *args, **kwargs):
        self.dirty = False
        super().__init__(*args, **kwargs)

    async def update(self, *args, **kwargs) -> bool:
        """ Update the modal on the server
        :return: True if the modal was updated, False otherwise
        """
        if not self.dirty:
            return False
        self.dirty = False
        await super().update()
        return True

    async def delete(self):
        """ Delete the modal from the server and all its children """
        await asyncio.gather(*[self._base_resource.parent.delete_component(rd_id)
                               for rd_id in self['properties']['reportIds']])
        await super().delete()

    def add_report(self, report: Report):
        """ Add report to the modal without saving it to the server
        :param report: report to add
        """
        if report['id'] in self['properties']['reportIds']:
            logger.warning(f"Report {report['id']} already in modal")
            return

        self['properties']['reportIds'].append(report['id'])
        self.dirty = True

    def remove_report(self, report: Report):
        """ Remove report from the modal without saving it to the server
        :param report: report to remove
        """

        if report['id'] not in self['properties']['reportIds']:
            logger.warning(f"Report {report['id']} not in modal")
            return

        self['properties']['reportIds'].remove(report['id'])
        self.dirty = True

    def has_report(self, report: Report):
        """ Check if report is in the modal
        :param report: report to check
        """
        return report['id'] in self['properties']['reportIds']

    def clear_content(self):
        """ Remove all reports from the tabs group without saving it to the server """
        self['properties']['reportIds'] = []
        self.dirty = True
